'''
Late Emergence of the Whisker Direction Selectivity Map in the Rat Barrel Cortex
Kremer Y, Leger JF, Goodman DF, Brette R, Bourdieu L (2011). J Neurosci 31(29):10689-700.

Development of direction maps with pinwheels in the barrel cortex.
Whiskers are deflected with random moving bars.
N.B.: network construction can be long.

In this version, STDP is faster than in the paper so that the script runs in just a
few minutes.
'''
from brian import *

# Uncomment if you have a C compiler
# set_global_preferences(useweave=True,usecodegen=True,usecodegenweave=True,usenewpropagate=True,usecstdp=True)

# PARAMETERS
# Neuron numbers
M4,M23exc,M23inh=22,25,12 # side of each barrel (in neurons)
N4,N23exc,N23inh=M4**2,M23exc**2,M23inh**2 # neurons per barrel
barrelarraysize=5 # Choose 3 or 4 if memory error
Nbarrels=barrelarraysize**2
# Stimulation
stim_change_time = 5*ms
Fmax=.5/stim_change_time # maximum firing rate in layer 4 (.5 spike / stimulation)
# Neuron parameters
taum,taue,taui=10*ms,2*ms,25*ms
El=-70*mV
Vt,vt_inc,tauvt=-55*mV,2*mV,50*ms # adaptive threshold
# STDP
taup,taud=5*ms,25*ms
Ap,Ad=.05,-.04
# EPSPs/IPSPs
EPSP,IPSP = 1*mV,-1*mV
EPSC = EPSP * (taue/taum)**(taum/(taue-taum))
IPSC = IPSP * (taui/taum)**(taum/(taui-taum))
# Model: IF with adaptive threshold
eqs='''
dv/dt=(ge+gi+El-v)/taum : volt
dge/dt=-ge/taue : volt
dgi/dt=-gi/taui : volt
dvt/dt=(Vt-vt)/tauvt : volt # adaptation
x : 1
y : 1
'''
# Tuning curve
tuning=lambda theta:clip(cos(theta),0,Inf)*Fmax

# Layer 4
layer4=PoissonGroup(N4*Nbarrels)
barrels4 = dict(((i, j), layer4.subgroup(N4)) for i in xrange(barrelarraysize) for j in xrange(barrelarraysize))
barrels4active = dict((ij, False) for ij in barrels4)
barrelindices = dict((ij, slice(b._origin, b._origin+len(b))) for ij, b in barrels4.iteritems())
layer4.selectivity = zeros(len(layer4))
for (i, j), inds in barrelindices.iteritems():
    layer4.selectivity[inds]=linspace(0,2*pi,N4)

# Layer 2/3
layer23=NeuronGroup(Nbarrels*(N23exc+N23inh),model=eqs,threshold='v>vt',reset='v=El;vt+=vt_inc',refractory=2*ms)
layer23.v=El
layer23.vt=Vt

# Layer 2/3 excitatory
layer23exc=layer23.subgroup(Nbarrels*N23exc)
x,y=meshgrid(arange(M23exc)*1./M23exc,arange(M23exc)*1./M23exc)
x,y=x.flatten(),y.flatten()
barrels23 = dict(((i, j), layer23exc.subgroup(N23exc)) for i in xrange(barrelarraysize) for j in xrange(barrelarraysize))
for i in range(barrelarraysize):
    for j in range(barrelarraysize):
        barrels23[i,j].x=x+i
        barrels23[i,j].y=y+j

# Layer 2/3 inhibitory
layer23inh=layer23.subgroup(Nbarrels*N23inh)
x,y=meshgrid(arange(M23inh)*1./M23inh,arange(M23inh)*1./M23inh)
x,y=x.flatten(),y.flatten()
barrels23inh = dict(((i, j), layer23inh.subgroup(N23inh)) for i in xrange(barrelarraysize) for j in xrange(barrelarraysize))
for i in range(barrelarraysize):
    for j in range(barrelarraysize):
        barrels23inh[i,j].x=x+i
        barrels23inh[i,j].y=y+j

print "Building synapses, please wait..."
# Feedforward connections
feedforward=Connection(layer4,layer23exc,'ge')
for i in range(barrelarraysize):
    for j in range(barrelarraysize):
        feedforward.connect_random(barrels4[i,j],barrels23[i,j],sparseness=.5,weight=EPSC*.5)
stdp=ExponentialSTDP(feedforward,taup,taud,Ap,Ad,wmax=EPSC)

# Excitatory lateral connections
recurrent_exc=Connection(layer23exc,layer23,'ge')
recurrent_exc.connect_random(layer23exc,layer23exc,weight=EPSC*.3,
                             sparseness=lambda i,j:.15*exp(-.5*(((layer23exc.x[i]-layer23exc.x[j])/.4)**2+((layer23exc.y[i]-layer23exc.y[j])/.4)**2)))
recurrent_exc.connect_random(layer23exc,layer23inh,weight=EPSC,
                             sparseness=lambda i,j:.15*exp(-.5*(((layer23exc.x[i]-layer23inh.x[j])/.4)**2+((layer23exc.y[i]-layer23inh.y[j])/.4)**2)))

# Inhibitory lateral connections
recurrent_inh=Connection(layer23inh,layer23exc,'gi')
recurrent_inh.connect_random(layer23inh,layer23exc,weight=IPSC,
                         sparseness=lambda i,j:exp(-.5*(((layer23inh.x[i]-layer23exc.x[j])/.2)**2+((layer23inh.y[i]-layer23exc.y[j])/.2)**2)))

# Stimulation
stimspeed = 1./stim_change_time # speed at which the bar of stimulation moves
direction = 0.0
stimzonecentre = ones(2)*barrelarraysize/2.
stimcentre,stimnorm = zeros(2),zeros(2)
stimradius = (11*stim_change_time*stimspeed+1)*.5
stimradius2 = stimradius**2

def new_direction():
    global direction
    direction = rand()*2*pi
    stimnorm[:] = (cos(direction), sin(direction))
    stimcentre[:] = stimzonecentre-stimnorm*stimradius

@network_operation
def stimulation():
    global direction, stimcentre
    stimcentre += stimspeed*stimnorm*defaultclock.dt
    if sum((stimcentre-stimzonecentre)**2)>stimradius2:
        new_direction()
    for (i, j), b in barrels4.iteritems():
        whiskerpos = array([i,j], dtype=float)+0.5
        isactive = abs(dot(whiskerpos-stimcentre, stimnorm))<.5
        if barrels4active[i, j]!=isactive:
            barrels4active[i, j] = isactive
            b.rate = float(isactive)*tuning(layer4.selectivity[barrelindices[i, j]]-direction)

new_direction()

run(5*second,report='text')

figure()
# Preferred direction
selectivity=array([mean(array(feedforward[:,i].todense())*exp(layer4.selectivity*1j)) for i in range(len(layer23exc))])
selectivity=(arctan2(selectivity.imag,selectivity.real) % (2*pi))*180./pi

I=zeros((barrelarraysize*M23exc,barrelarraysize*M23exc))
ix=array(around(layer23exc.x*M23exc),dtype=int)
iy=array(around(layer23exc.y*M23exc),dtype=int)
I[iy,ix]=selectivity
imshow(I)
hsv()
colorbar()
for i in range(1,barrelarraysize+1):
    plot([i*max(ix)/barrelarraysize,i*max(ix)/barrelarraysize],[0,max(iy)],'k')
    plot([0,max(ix)],[i*max(iy)/barrelarraysize,i*max(iy)/barrelarraysize],'k')

figure()
hist(selectivity)

show()



'''
Example of a 2D stimulus, see the
`complete description <http://neuralensemble.org/cookbook/wiki/Brian/StimulusArrayGroup>`__
at the Brian Cookbook.
'''

from brian import *
import scipy.ndimage as im

__all__ = ['bar', 'StimulusArrayGroup']

def bar(width, height, thickness, angle):
    '''
    An array of given dimensions with a bar of given thickness and angle
    '''
    stimulus = zeros((width, height))
    stimulus[:, int(height / 2. - thickness / 2.):int(height / 2. + thickness / 2.)] = 1.
    stimulus = im.rotate(stimulus, angle, reshape=False)
    return stimulus


class StimulusArrayGroup(PoissonGroup):
    '''
    A group of neurons which fire with a given stimulus at a given rate
    
    The argument ``stimulus`` should be a 2D array with values between 0 and 1.
    The point in the stimulus array at position (y,x) will correspond to the
    neuron with index i=y*width+x. This neuron will fire Poisson spikes at
    ``rate*stimulus[y,x]`` Hz. The stimulus will start at time ``onset``
    for ``duration``.
    '''
    def __init__(self, stimulus, rate, onset, duration):
        height, width = stimulus.shape
        stim = stimulus.ravel()*rate
        self.stimulus = stim
        def stimfunc(t):
            if onset < t < (onset + duration):
                return stim
            else:
                return 0. * Hz
        PoissonGroup.__init__(self, width * height, stimfunc)

if __name__ == '__main__':
    import pylab
    subplot(121)
    stim = bar(100, 100, 10, 90) * 0.9 + 0.1
    pylab.imshow(stim, origin='lower')
    pylab.gray()
    G = StimulusArrayGroup(stim, 50 * Hz, 100 * ms, 100 * ms)
    M = SpikeMonitor(G)
    run(300 * ms)
    subplot(122)
    raster_plot(M)
    axis(xmin=0, xmax=300)
    show()



from numpy import *
import multiprocessing
import time
from multiprocessing import sharedctypes
from numpy import ctypeslib
import ctypes

def f(x, a, n):
    s = 0
    for _ in xrange(n):
        s += sum(x) * a
    return s

def process_listener(x, conn):
    x = ctypeslib.as_array(x)
    while True:
        conn.poll()
        a, n = conn.recv()
        conn.send(f(x, a, n))

if __name__ == '__main__':
    N = 10000000
    numprocesses = 6
    complexity = 10

    print 'Array size:', N
    print 'Num processors:', numprocesses
    print 'Task complexity:', complexity
    print

    x = ones(N)
    x_sct = sharedctypes.Array(ctypes.c_double, x, lock=False)
    A = 1 + arange(numprocesses)

    pipes = [multiprocessing.Pipe() for _ in xrange(numprocesses)]
    server_conns, client_conns = zip(*pipes)
    processes = [multiprocessing.Process(
                            target=process_listener,
                            args=(x_sct, conn)
                            ) for conn in client_conns]
    for p in processes:
        p.start()

    start = time.time()
    for i, a in enumerate(A):
        server_conns[i].send((a, complexity))
    for i in xrange(numprocesses):
        server_conns[i].poll()
        server_conns[i].recv()
    end = time.time()
    t_multiprocessing_1 = end - start

    start = time.time()
    for i, a in enumerate(A):
        server_conns[i].send((a, complexity))
    for i in xrange(numprocesses):
        server_conns[i].poll()
        server_conns[i].recv()
    end = time.time()
    t_multiprocessing_2 = end - start

    for p in processes:
        p.terminate()

    start = time.time()
    for a in A:
        f(x, a, complexity)
    end = time.time()
    t_single = end - start

    print 'Single:', t_single
    print 'Multiprocessing 1:', t_multiprocessing_1
    print 'Multiprocessing 2:', t_multiprocessing_2
    print 'S/M1:', t_single / t_multiprocessing_1
    print 'S/M2:', t_single / t_multiprocessing_2



'''
Phase locking of IF neurons to a periodic input
'''
from brian import *

tau = 20 * ms
N = 100
b = 1.2 # constant current mean, the modulation varies
f = 10 * Hz

eqs = '''
dv/dt=(-v+a*sin(2*pi*f*t)+b)/tau : 1
a : 1
'''

neurons = NeuronGroup(N, model=eqs, threshold=1, reset=0)
neurons.v = rand(N)
neurons.a = linspace(.05, 0.75, N)
S = SpikeMonitor(neurons)
trace = StateMonitor(neurons, 'v', record=50)

run(1000 * ms)
subplot(211)
raster_plot(S)
subplot(212)
plot(trace.times / ms, trace[50])
show()



# ----------------------------------------------------------------------------------
# Copyright ENS, INRIA, CNRS
# Contributors: Romain Brette (brette@di.ens.fr) and Dan Goodman (goodman@di.ens.fr)
# 
# Brian is a computer program whose purpose is to simulate models
# of biological neural networks.
# 
# This software is governed by the CeCILL license under French law and
# abiding by the rules of distribution of free software.  You can  use, 
# modify and/ or redistribute the software under the terms of the CeCILL
# license as circulated by CEA, CNRS and INRIA at the following URL
# "http://www.cecill.info". 
# 
# As a counterpart to the access to the source code and  rights to copy,
# modify and redistribute granted by the license, users are provided only
# with a limited warranty  and the software's author,  the holder of the
# economic rights,  and the successive licensors  have only  limited
# liability. 
# 
# In this respect, the user's attention is drawn to the risks associated
# with loading,  using,  modifying and/or developing or reproducing the
# software by the user in light of its specific status of free software,
# that may mean  that it is complicated to manipulate,  and  that  also
# therefore means  that it is reserved for developers  and  experienced
# professionals having in-depth computer knowledge. Users are therefore
# encouraged to load and test the software's suitability as regards their
# requirements in conditions enabling the security of their systems and/or 
# data to be ensured and,  more generally, to use and operate it in the 
# same conditions as regards security. 
# 
# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.
# ----------------------------------------------------------------------------------
# 
'''
Inspection of strings with python statements 
and models defined by differential equations.

TODO: some of the module is obsolete
'''

__all__ = ['is_affine', 'depends_on', 'Term', 'get_global_term',
         'get_var_names', 'check_equations_units', 'fill_vars', 'AffineFunction',
         'get_identifiers', 'modified_variables', 'namespace', 'clean_text',
         'namespace_replace_quantity_with_pure', 'list_replace_quantity_with_pure']

import numpy
from numpy import array
from units import *
import parser
import re
import inspect
from copy import copy


class PureQuantityBase(Quantity):
    def __init__(self, value):
        self.dim = value.dim

    def __div__(self, other):
        if not isinstance(other, Quantity) and not is_scalar_type(other):
            return NotImplemented
        try:
            return Quantity.__div__(self, other)
        except ZeroDivisionError:
            try:
                odim = other.dim
            except AttributeError:
                odim = Dimension()
            return Quantity.with_dimensions(0, self.dim / odim)
    __truediv__ = __div__
    def __rdiv__(self, other):
        if not isinstance(other, Quantity) and not is_scalar_type(other):
            return NotImplemented
        try:
            return Quantity.__rdiv__(self, other)
        except ZeroDivisionError:
            try:
                odim = other.dim
            except AttributeError:
                odim = Dimension()
            return Quantity.with_dimensions(0, odim / self.dim)
    __rtruediv__ = __rdiv__
    def __mod__(self, other):
        if not isinstance(other, Quantity) and not is_scalar_type(other):
            return NotImplemented
        try:
            return Quantity.__mod__(self, other)
        except ZeroDivisionError:
            return Quantity.with_dimensions(0, self.dim)

def returnpure(meth):
    def f(*args, **kwds):
        x = meth(*args, **kwds)
        if isinstance(x, Quantity):
            return PureQuantity(x)
        else:
            return x
    return f


class PureQuantity(PureQuantityBase):
    '''
    Use this class for unit checking.
    
    The idea is that operations should always work if they are
    dimensionally consistent regardless of the values. The key one is that
    division by zero does not raise an error but returns a value with the
    correct dimensions (in fact, it returns 0 with the correct dimensions).
    This is important for Brian because we do unit checking by substituting
    zeros into the equations, which sometimes gives a divide by zero error.
    
    The way it works is that it derives from Quantity, but for division it
    wraps a try: except ZeroDivisionError: around the operation, and returns
    a zero with the correct units if it encounters it. In addition to that,
    it wraps every method of Quantity so that they return PureQuantity objects
    instead of Quantity objects (otherwise e.g. a*b would be a Quantity not
    a PureQuantity even if a, b were PureQuantity).
    
    Finally, the *_replace_quantity_with_pure functions are just designed to
    scan through a dict or list of variables and replace Quantity objects
    by PureQuantity objects. You need to do this when evaluating some code
    in a user namespace, for example.
    '''
    for methname in dir(PureQuantityBase):
        meth = getattr(PureQuantityBase, methname)
        try:
            meth2 = getattr(numpy.float64, methname)
        except AttributeError:
            meth2 = meth
        if callable(meth) and meth is not meth2:
            exec methname + '=returnpure(PureQuantityBase.' + methname + ')'
    del meth, meth2, methname

def namespace_replace_quantity_with_pure(ns):
    newns = {}
    for k, v in ns.iteritems():
        if isinstance(v, Quantity):
            v = PureQuantity(v)
        newns[k] = v
    return newns

def list_replace_quantity_with_pure(L):
    newL = []
    for v in L:
        if isinstance(v, Quantity):
            v = PureQuantity(v)
        newL.append(v)
    return newL

def namespace(expr, level=0, return_unknowns=False):
    '''
    Returns a namespace with the values of identifiers in expr,
    taking from:
    * local namespace
    * global namespace
    * units
    '''
    # Build the namespace
    frame = inspect.stack()[level + 1][0]
    global_namespace, local_namespace = frame.f_globals, frame.f_locals
    # Find external objects
    space = {}
    unknowns = []
    for var in get_identifiers(expr):
        if var in local_namespace: #local
            space[var] = local_namespace[var]
        elif var in global_namespace: #global
            space[var] = global_namespace[var]
        elif var in globals(): # typically units
            space[var] = globals()[var]
        else:
            unknowns.append(var)
    if return_unknowns:
        return space, unknowns
    else:
        return space

def get_identifiers(expr):
    '''
    Returns the list of identifiers (variables or functions) in the
    Python expression (string).
    '''
    # cleaner: parser.expr(expr).tolist() then find leaves of the form [1,name]
    return parser.suite(expr).compile().co_names

def clean_text(expr):
    '''
    Cleans a Python expression or statement:
    * Remove comments (# comment)
    * Merge multi-line statements (\)
    * Split at semi-columns (careful: indentation is ignored)
    '''
    # Merge multi-line statements
    expr = re.sub('\\\s*?\n', ' ', expr)
    # Remove comments
    expr = re.sub('#.*', '', expr)
    # Split at semi-columns
    expr = re.sub(';', '\n', expr)
    return expr

def modified_variables(expr):
    '''
    Returns the list of variables or functions in expr that are in left-hand sides, e.g.:
        x+=5
    expr can be a multiline statement.
    Multiline comments are not allowed but multiline statements are.
    Functions may also be returned as in:
        do(something) # here do is returned, not something
        
    TODO: maybe functions should be removed?
    TODO: better handling of semi-columns (;)
    '''
    vars = get_identifiers(expr)
    expr = clean_text(expr)
    # Find lines that start by an identifier
    mod_vars = []
    for line in expr.splitlines():
        s = re.search(r'^\s*(\w+)\b', line)
        if s and (s.group(1) in vars):
            mod_vars.append(s.group(1))
    return mod_vars

def fill_vars(f, keepnamespace=False, *varnames):
    '''
    Returns a function with arguments given by varnames (list or tuple),
    given that the arguments of f are in varnames.
    If keepnamespace is True, then the original func_globals dictionary of
    the function is kept.
    Purpose: changing the syntax for function calls.
    Example:
    f=lambda x:2*x
    g=fill_vars(f,'y','x')
    Then g(1,2) returns f(2).
    This is somehow the inverse of partial (in module functools).
    N.B.: the order of variables matters.
    '''
    if list(f.func_code.co_varnames) == varnames:
        return f
    varstring = varnames[0]
    for name in varnames[1:]:
        varstring += ',' + name
    shortvarstring = f.func_code.co_varnames[0]
    for name in f.func_code.co_varnames[1:]:
        shortvarstring += ',' + name
    if keepnamespace:
        # Create a unique name
        fname = 'fill_vars_function' + id(f)
        f.func_globals[fname] = f
        return eval('lambda ' + varstring + ': ' + fname + '(' + shortvarstring + ')', f.func_globals)
    else:
        return eval('lambda ' + varstring + ': f(' + shortvarstring + ')', {'f':f})

def check_equations_units(eqs, x):
    '''
    Check the units of the differential equations, using
    the units of x.
    df_i/dt must have units of x_i / time.
    '''
    try:
        for f, x_i in zip(eqs, x):
            f.func_globals['xi'] = 0 * second ** -.5 # Noise
            f(*x) + (x_i / second) # Check that the two terms have the same dimension
    except DimensionMismatchError, inst:
        raise DimensionMismatchError("The differential equations are not homogeneous!", *inst._dims)

def get_var_names(eqs):
    '''
    Get the variable names from the set of equations.
    Returns a list.
    N.B.: the order is preserved.
    '''
    names = list(eqs[0].func_code.co_varnames)
    for eq in eqs:
        for name in eq.func_code.co_varnames:
            if not(name in names):
                names.append(name)
    return names


class AffineFunction(object):
    '''
    An object that can be added and multiplied by a float (or array or int).
    '''
    def __init__(self, a=1., b=0.):
        '''
        Defines an affine function as a*x+b.
        '''
        self.a = a
        self.b = b

    def __add__(self, y):
        if isinstance(y, AffineFunction):
            return AffineFunction(self.a + y.a, self.b + y.b)
        else:
            return AffineFunction(self.a, self.b + array(y))

    def __radd__(self, x):
        if isinstance(x, AffineFunction):
            return AffineFunction(self.a + x.a, self.b + x.b)
        else:
            return AffineFunction(self.a, self.b + array(x))

    def __neg__(self):
        return AffineFunction(-self.a, -self.b)

    def __sub__(self, y):
        if isinstance(y, AffineFunction):
            return AffineFunction(self.a - y.a, self.b - y.b)
        else:
            return AffineFunction(self.a, self.b - array(y))

    def __rsub__(self, x):
        if isinstance(x, AffineFunction):
            return AffineFunction(x.a - self.a, x.b - self.b)
        else:
            return AffineFunction(-self.a, array(x) - self.b)

    def __mul__(self, y):
        if isinstance(y, float) or isinstance(y, int) or isinstance(y, array):
            return AffineFunction(self.a * array(y), self.b * array(y))
        else:
            return y.__rmul__(self)

    def __rmul__(self, x):
        if isinstance(x, float) or isinstance(x, int) or isinstance(x, array):
            return AffineFunction(array(x) * self.a, array(x) * self.b)
        else:
            return x.__mul__(self)

    def __div__(self, y):
        if isinstance(y, float) or isinstance(y, int) or isinstance(y, array):
            return AffineFunction(self.a / array(y), self.b / array(y))
        else:
            return y.__rdiv__(self)

    def __repr__(self):
        return str(self.a) + '*x+' + str(self.b)


class Term(object):
    '''
    A variable that can be used to isolate terms in
    a function, e.g.:
    f=lambda x:a+3*x
    f(Term()) --> 3*Term()
    Idea:
      a+x(z) = x(z) + a = x(z)
      a*x(z) = x(z)*a = x(a*z)
    '''
    def __init__(self, x=1.):
        self.x = x

    def __add__(self, y):
        return self

    def __radd__(self, y):
        return self

    def __mul__(self, y):
        return Term(self.x * y)

    def __rmul__(self, y):
        return Term(self.x * y)

    def __div__(self, y):
        return Term(self.x / y)

    def __neg__(self):
        return Term(-self.x)

    def __repr__(self):
        return str(self.x) + '*Term()'

    def __print__(self):
        return str(self.x) + '*Term()'

def is_affine(f):
    '''
    Tests whether f is an affine function.
    '''
    nargs = f.func_code.co_argcount
    try:
        f(*([AffineFunction()]*nargs))
    except:
        return False
    return True

#def is_affine1st(f,x0):
#    '''
#    Tests whether f is affine in its 1st variable.
#    '''
#    nargs=f.func_code.co_argcount
#    return is_affine(lambda x:f(x,*x0))

def depends_on(f, x, x0):
    '''
    Tests whether f depends on global variable x.
    N.B.: returns True also if f generates an error
    (e.g. with undefined variables).
    x0 is the test value for the variables (tuple).
    
    Other idea: use 'xi' in f.func_code.co_names (but not working for nested functions)
    '''
    #return x in f.func_code.co_names
#    if x in f.func_globals:
#        oldx=f.func_globals[x]
#    else:
#        oldx=None

    old_func_globals = copy(f.func_globals)

    x0 = list_replace_quantity_with_pure(x0)
    f.func_globals.update(namespace_replace_quantity_with_pure(f.func_globals))

    result = False
    f.func_globals[x] = None
    nargs = f.func_code.co_argcount
    try:
        f(*x0)
    except:
        result = True

#    if oldx==None:
#        del f.func_globals[x] # x was not defined
#    else:
#        f.func_globals[x]=oldx # previous value

    f.func_globals.update(old_func_globals)

    return result

def get_global_term(f, x, x0):
    '''
    Extract the term in global variable x from function x,
    returns a float.
    x0 is the test value for the variables (tuple).
    Example:
    getterm(lambda x:2*x+5*y,'y',(0,0))) --> 5
    '''
    old_func_globals = copy(f.func_globals)

    x0 = list_replace_quantity_with_pure(x0)
    f.func_globals.update(namespace_replace_quantity_with_pure(f.func_globals))

    f.func_globals[x] = Term()

    result = f(*x0).x

    f.func_globals.update(old_func_globals)

    return result



from brian import *
set_global_preferences(usenewbrianhears=True,
                       #useweave=False,
                       )
from brian.hears import *
from scipy.signal import lfilter
import time

N = 1000
duration = 10240
bufsize = 1024

x = Sound(randn(duration, 1), samplerate=44100*Hz)

fb = GammatoneFilterbank(x, erbspace(20*Hz, 20*kHz, N))

if 0:
    # Doing it with lfilter channel by channel
    b, a = fb.filt_b, fb.filt_a
    n, m, p = b.shape
    start = time.time()
    for i in xrange(n):
        for k in xrange(p):
            lfilter(b[i,:,k], a[i,:,k], x)
    print time.time()-start

fb.buffer_init()
start = time.time()
for i in xrange(0, duration, bufsize):
    fb.buffer_fetch(i, i+bufsize)
end = time.time()

print end-start



# NOTE: this is now in Brian.connections

from brian import *
from brian.connection import ConnectionMatrix
import random as pyrandom
from scipy import random as scirandom
import numpy

__all__ = ['UserComputedConnectionMatrix', 'UserComputedSparseConnectionMatrix', 'random_row_func', 'random_sparse_row_func']

def random_row_func(N, p, weight=1., initseed=None):
    '''
    Returns a random connectivity ``row_func`` for use with :class:`UserComputedConnectionMatrix`
    
    Gives equivalent output to the :meth:`Connection.connect_random` method.
    
    Arguments:
    
    ``N``
        The number of target neurons.
    ``p``
        The probability of a synapse.
    ``weight``
        The connection weight (must be a single value).
    ``initseed``
        The initial seed value (for reproducible results).
    '''
    if initseed is None:
        initseed = pyrandom.randint(100000, 1000000) # replace this
    cur_row = numpy.zeros(N)
    myrange = numpy.arange(N, dtype=int)

    def row_func(i):
        pyrandom.seed(initseed + int(i))
        scirandom.seed(initseed + int(i))
        k = scirandom.binomial(N, p, 1)[0]
        cur_row[:] = 0.0
        cur_row[pyrandom.sample(myrange, k)] = weight
        return cur_row
    return row_func


class UserComputedConnectionMatrix(ConnectionMatrix):
    '''
    A computed connection matrix defined by a user-specified function
    
    Normally this matrix will be initialised by passing the class
    object to the :class:`Connection` object. In the initialisation
    of the :class:`Connection` specify ``structure=UserComputedConnectionMatrix``
    and add the keyword ``row_func=...``, e.g.::
    
        def f(i):
            return max_weight*ones(N)/(1+(arange(N)-i)**2)
        C = Connection(G1, G2, structure=UserComputedConnectionMatrix, row_func=f)
    
    Initialisation arguments:
    
    ``dims``
        The pair ``(N,M)`` specifying the dimensions of the matrix.
    ``row_func``
        The function ``f(i)`` which returns an array of length ``M``,
        the weight matrix for row ``i``. Note that you are responsible
        for making sure the function returns consistent results (so
        random functions should be initialised with a seed based on
        the row ``i``).
    
    **Limitations**
    
    This type of connection matrix cannot be changed during a run, and
    cannot be used with methods like :class:`Connection.connect_random`.
    
    **Efficiency considerations**
    
    This connection matrix is for dense connectivity, if the connectivity
    is sparse you might get better performance with :class:`UserComputedSparseConnectionMatrix`.
    '''
    def __init__(self, dims, row_func):
        self.sourcelen, self.targetlen = dims
        self.row_func = row_func

    def get_row(self, i):
        return self.row_func(i)

    def __getitem__(self, item):
        if isinstance(item, int):
            return self.get_row(item)
        if isinstance(item, tuple):
            if len(item) == 2:
                item_i, item_j = item
                if isinstance(item_i, int) and isinstance(item_j, slice):
                    if is_colon_slice(item_j):
                        return self.get_row(item_i)
        raise ValueError('Only "i,:" indexing supported.')

def random_sparse_row_func(N, p, weight=1., initseed=None):
    '''
    Returns a random connectivity ``row_func`` for use with :class:`UserComputedSparseConnectionMatrix`
    
    Gives equivalent output to the :meth:`Connection.connect_random` method.
    
    Arguments:
    
    ``N``
        The number of target neurons.
    ``p``
        The probability of a synapse.
    ``weight``
        The connection weight (must be a single value).
    ``initseed``
        The initial seed value (for reproducible results).
    '''
    if initseed is None:
        initseed = pyrandom.randint(100000, 1000000) # replace this
    myrange = numpy.arange(N, dtype=int)
    def row_func(i):
        pyrandom.seed(initseed + int(i))
        scirandom.seed(initseed + int(i))
        k = scirandom.binomial(N, p, 1)[0]
        return (pyrandom.sample(myrange, k), weight)
    return row_func


class UserComputedSparseConnectionMatrix(ConnectionMatrix):
    '''
    A computed sparse connection matrix defined by a user-specified function
    
    Normally this matrix will be initialised by passing the class
    object to the :class:`Connection` object. In the initialisation
    of the :class:`Connection` specify ``structure=UserComputedSparseConnectionMatrix``
    and add the keyword ``row_func=...``, e.g.::
    
        def f(i):
            if 0<i<N-1:
                return ([i-1,i+1], weight*ones(2))
            elif i>0:
                return ([i-1], weight*ones(1))
            else:
                return ([i+1], weight*ones(1))
        C = Connection(G1, G2, structure=UserComputedSparseConnectionMatrix, row_func=f)
    
    Initialisation arguments:
    
    ``dims``
        The pair ``(N,M)`` specifying the dimensions of the matrix.
    ``row_func``
        The function ``f(i)`` which for a row ``i`` returns a pair ``(indices, values))``
        consisting of a list or array ``indices`` with the indices of the
        nonzero elements of the row, and an array of the same length ``values``
        giving the weight matrix for those indices. Note that you are responsible
        for making sure the function returns consistent results (so
        random functions should be initialised with a seed based on
        the row ``i``).
    
    **Limitations**
    
    This type of connection matrix cannot be changed during a run, and
    cannot be used with methods like :class:`Connection.connect_random`.
    
    **Efficiency considerations**
    
    This connection matrix is for sparse connectivity, if the connectivity
    is dense you might get better performance with :class:`UserComputedConnectionMatrix`.
    '''
    def __init__(self, dims, row_func):
        self.sourcelen, self.targetlen = dims
        self.row_func = row_func
        self.cur_row = numpy.zeros(dims[1])

    def add_row(self, i, X):
        indices, values = self.row_func(i)
        X[indices] += values

    def add_scaled_row(self, i, X, factor):
        # modulation may not work? need factor[self.rows[i]] here? is factor a number or an array?
        X[indices] += factor * values

    def get_row(self, i):
        indices, values = self.row_func(i)
        self.cur_row[:] = 0.0
        self.cur_row[indices] = values
        return self.cur_row

    def __getitem__(self, item):
        if isinstance(item, int):
            return self.get_row(item)
        if isinstance(item, tuple):
            if len(item) == 2:
                item_i, item_j = item
                if isinstance(item_i, int) and isinstance(item_j, slice):
                    if is_colon_slice(item_j):
                        return self.get_row(item_i)
        raise ValueError('Only "i,:" indexing supported.')

if __name__ == '__main__':

    eqs = '''
    dv/dt = (ge+gi-(v+49*mV))/(20*ms) : volt
    dge/dt = -ge/(5*ms) : volt
    dgi/dt = -gi/(10*ms) : volt
    '''

    N = 4000
    #N = 20000
    #N = 100000
    #N = 1000000
    duration = 1000 * ms
    usecc = True
    usesparsecc = True

    Ne = int(0.8 * N)
    Ni = N - Ne
    prob = 80. / N

    P = NeuronGroup(N, model=eqs, threshold= -50 * mV, reset= -60 * mV)
    P.v = -60 * mV + 10 * mV * rand(len(P))
    Pe = P.subgroup(Ne)
    Pi = P.subgroup(Ni)

    if usecc:
        if usesparsecc:
            connmat = UserComputedSparseConnectionMatrix
            rfunc = random_sparse_row_func
        else:
            connmat = UserComputedConnectionMatrix
            rfunc = random_row_func
        Ce = Connection(Pe, P, 'ge', structure=connmat,
                      row_func=rfunc(N, p=prob, weight=1.62 * mV))
        Ci = Connection(Pi, P, 'gi', structure=connmat,
                      row_func=rfunc(N, p=prob, weight= -9 * mV))
    else:
        Ce = Connection(Pe, P, 'ge')
        Ci = Connection(Pi, P, 'gi')
        Ce.connect_random(Pe, P, prob, weight=1.62 * mV)
        Ci.connect_random(Pi, P, prob, weight= -9 * mV)

    if N > 5000:
        M = PopulationSpikeCounter(P)
    else:
        M = SpikeMonitor(P)

    if N > 100000:
        @network_operation()
        def f(t):
            print 'Current time', t

    import time
    t = time.time()
    run(duration)
    if usecc:
        if usesparsecc:
            print 'Using sparse computed connection'
        else:
            print 'Using computed connection'
    else:
        print 'Using standard connection'
    print "Time taken to run simulation:", time.time() - t
    print "Number of spikes:", M.nspikes
    if N <= 5000:
        raster_plot(M)
        show()

# Statistics, sample runs on my laptop
#
#N=4000, T=1s
#
#Using standard connection
#Time taken to run simulation: 5.76499986649
#Number of spikes: 24860
#
#Using computed connection
#Time taken to run simulation: 12.0629999638
#Number of spikes: 24761
#
#Using sparse computed connection
#Time taken to run simulation: 11.3589999676
#Number of spikes: 23421
#
#N=20000, T=1s
#
#Using standard connection
#Time taken to run simulation: 23.8900001049
#Number of spikes: 122241
#RAM usage: 94MB
#
#Using computed connection
#Time taken to run simulation: 74.75
#Number of spikes: 121995
#RAM usage: 39MB
#
#Using sparse computed connection
#Time taken to run simulation: 49.3440001011
#Number of spikes: 121841
#RAM usage: 39MB
#
#N=100000, T=10ms
#
#Using standard connection
#Time taken to run simulation: 8.17100000381
#Number of spikes: 10971
#RAM usage: 330MB
#
#Using computed connection
#Time taken to run simulation: 13.390999794
#Number of spikes: 11299
#RAM usage: 45MB
#
#Using sparse computed connection
#Time taken to run simulation: 4.98399996758
#Number of spikes: 11035
#RAM usage: 45MB
#
#N=1m, T=1ms
#
#Using computed connection
#Time taken to run simulation: 137.796999931
#Number of spikes: 6219
#RAM usage: 95MB
#
#Using sparse computed connection
#Time taken to run simulation: 4.04700016975
#Number of spikes: 6181
#RAM usage: 120MB



'''
Adapted from
Irregular persistent activity induced by synaptic excitatory feedback
F Barbieri, N Brunel - Frontiers in Computational Neuroscience, 2007

R Brette 2009
'''
from brian import *

theta = 20 * mV
Vr = 15 * mV
tau = 5 * ms
tau_rp = 2 * ms
N = 800
nu_sp = 3 * Hz
mu_ext = 10 * mV # (?)
sigma_ext = 8 * mV
tau_r = .05 * ms
tau_rn = 2 * ms
tau_d = 5 * ms
tau_dn = 100 * ms
gamma = .9
u = 0.5
tau_rec = 160 * ms
lambd = 0.1
D = 1 * ms # delay (?)
J = 35 * mV # (?)

#defaultclock.dt=.001*ms

eqs = '''
dv/dt=(-v+I_rec+mu_ext*tau**.5*xi)/tau : volt
I_rec=s+z : volt
ds/dt=(x-s)/tau_d : volt
dx/dt=-x/tau_r : volt
dz/dt=(h-z)/tau_dn : volt
dh/dt=-h/tau_rn : volt
'''

neurons = NeuronGroup(N, model=eqs, threshold=theta, reset=Vr, refractory=tau_rp)
synapses_ampa = Connection(neurons, neurons, 'x', weight=(1 - gamma) * tau / (tau_r * N) * J)
synapses_nmda = Connection(neurons, neurons, 'h', weight=gamma * tau / (tau_rn * N) * J)
stp1 = STP(synapses_ampa, taud=tau_rec, tauf=.1 * ms, U=u)
stp2 = STP(synapses_nmda, taud=tau_rec, tauf=.1 * ms, U=u)

spikes = SpikeMonitor(neurons)

run(1 * second)
#run(500*ms)
raster_plot(spikes)
print spikes.nspikes * 1. / N
show()



# ###########################################
#
# Inhibitory synaptic plasticity in a recurrent network model 
# (F. Zenke, 2011)
#
# Adapted from: 
# Vogels, T. P., H. Sprekeler, F. Zenke, C. Clopath, and W. Gerstner. 
# 'Inhibitory Plasticity Balances Excitation and Inhibition in Sensory
# Pathways and Memory Networks.' Science (November 10, 2011). 
#
# ###########################################

from brian import *

# ###########################################
# Defining network model parameters
# ###########################################

NE = 8000          # Number of excitatory cells
NI = NE/4          # Number of inhibitory cells 

w = 1.*nS           # Basic weight unit
tau_ampa = 5.0*ms   # Glutamatergic synaptic time constant
tau_gaba = 10.0*ms  # GABAergic synaptic time constant
epsilon = 0.02      # Sparseness of synaptic connections

eta = 1e-2          # Learning rate
tau_stdp = 20*ms    # STDP time constant

simtime = 10*second # Simulation time

# ###########################################
# Neuron model
# ###########################################

gl = 10.0*nsiemens   # Leak conductance
el = -60*mV          # Resting potential
er = -80*mV          # Inhibitory reversal potential
vt = -50.*mV         # Spiking threshold
memc = 200.0*pfarad  # Membrane capacitance
bgcurrent = 200*pA   # External current

eqs_neurons='''
dv/dt=(-gl*(v-el)-(g_ampa*w*v+g_gaba*(v-er)*w)+bgcurrent)/memc : volt
dg_ampa/dt = -g_ampa/tau_ampa : 1
dg_gaba/dt = -g_gaba/tau_gaba : 1
'''

# ###########################################
# Initialize neuron group
# ###########################################

neurons=NeuronGroup(NE+NI,model=eqs_neurons,threshold=vt,reset=el,refractory=5*ms)
Pe=neurons.subgroup(NE)
Pi=neurons.subgroup(NI)

# ###########################################
# Connecting the network 
# ###########################################

con_e = Connection(Pe,neurons,'g_ampa',weight=0.3,sparseness=epsilon)
con_ie = Connection(Pi,Pe,'g_gaba',weight=1e-10,sparseness=epsilon)
con_ii = Connection(Pi,Pi,'g_gaba',weight=3,sparseness=epsilon)

# ###########################################
# Setting up monitors
# ###########################################

sm = SpikeMonitor(Pe)

# ###########################################
# Run without plasticity
# ###########################################

run(1*second)

# ###########################################
# Inhibitory Plasticity
# ###########################################

alpha = 3*Hz*tau_stdp*2  # Target rate parameter
gmax = 100               # Maximum inhibitory weight

eqs_stdp_inhib = '''
dA_pre/dt=-A_pre/tau_stdp : 1
dA_post/dt=-A_post/tau_stdp : 1
'''

stdp_ie = STDP(con_ie, eqs=eqs_stdp_inhib, pre='A_pre+=1.; w+=(A_post-alpha)*eta;',
        post='A_post+=1.; w+=A_pre*eta;', wmax=gmax)

# ###########################################
# Run with plasticity
# ###########################################

run(simtime-1*second,report='text')

# ###########################################
# Make plots
# ###########################################

subplot(211)
raster_plot(sm,ms=1.)
title("Before")
xlabel("")
xlim(0.8*1e3,1*1e3)
subplot(212)
raster_plot(sm,ms=1.)
title("After")
xlim((simtime-0.2*second)*1e3,simtime*1e3)
show()



'''
A rate model
'''
from brian import *

N = 50000
tau = 20 * ms
I = 10 * Hz
eqs = '''
dv/dt=(I-v)/tau : Hz # note the unit here: this is the output rate
'''
group = NeuronGroup(N, eqs, threshold=PoissonThreshold())
S = PopulationRateMonitor(group, bin=1 * ms)

run(100 * ms)

plot(S.rate)
show()



from brian import *

set_global_preferences(useweave=False)

G = NeuronGroup(3, 'V:1')
H = NeuronGroup(4, 'V:1')
C = Connection(G, H, 'V')
C.connect_random(p=0.5, weight=1)

C.compress()

print C.W.todense()
for i in range(C.W.shape[0]):
    print C.W[i, :].todense()
print
print C.W.todense().T
for i in range(C.W.shape[1]):
    print C.W[:, i].todense()



'''
Example of using Python multiprocessing module to distribute simulations over
multiple processors.

The general procedure for using multiprocessing is to define and run a network
inside a function, and then use multiprocessing.Pool.map to call the function
with multiple parameter values. Note that on Windows, any code that should only
run once should be placed inside an if __name__=='__main__' block.
'''

from brian import *
import multiprocessing

# This is the function that we want to compute for various different parameters
def how_many_spikes(excitatory_weight):
    # These two lines reset the clock to 0 and clear any remaining data so that
    # memory use doesn't build up over multiple runs.
    reinit_default_clock()
    clear(True)
    eqs = '''
    dv/dt = (ge+gi-(v+49*mV))/(20*ms) : volt
    dge/dt = -ge/(5*ms) : volt
    dgi/dt = -gi/(10*ms) : volt
    '''
    P = NeuronGroup(4000, eqs, threshold= -50 * mV, reset= -60 * mV)
    P.v = -60 * mV + 10 * mV * rand(len(P))
    Pe = P.subgroup(3200)
    Pi = P.subgroup(800)
    Ce = Connection(Pe, P, 'ge')
    Ci = Connection(Pi, P, 'gi')
    Ce.connect_random(Pe, P, 0.02, weight=excitatory_weight)
    Ci.connect_random(Pi, P, 0.02, weight= -9 * mV)
    M = SpikeMonitor(P)
    run(100 * ms)
    return M.nspikes

if __name__ == '__main__':
    # Note that on Windows platforms, all code that is executed rather than
    # just defining functions and classes has to be in the if __name__=='__main__'
    # block, otherwise it will be executed by each process that starts. This
    # isn't a problem on Linux.
    pool = multiprocessing.Pool() # uses num_cpu processes by default
    weights = linspace(0, 3.5, 100) * mV
    args = [w * volt for w in weights]
    results = pool.map(how_many_spikes, args) # launches multiple processes
    plot(weights, results, '.')
    show()



# ----------------------------------------------------------------------------------
# Copyright ENS, INRIA, CNRS
# Contributors: Romain Brette (brette@di.ens.fr) and Dan Goodman (goodman@di.ens.fr)
# 
# Brian is a computer program whose purpose is to simulate models
# of biological neural networks.
# 
# This software is governed by the CeCILL license under French law and
# abiding by the rules of distribution of free software.  You can  use, 
# modify and/ or redistribute the software under the terms of the CeCILL
# license as circulated by CEA, CNRS and INRIA at the following URL
# "http://www.cecill.info". 
# 
# As a counterpart to the access to the source code and  rights to copy,
# modify and redistribute granted by the license, users are provided only
# with a limited warranty  and the software's author,  the holder of the
# economic rights,  and the successive licensors  have only  limited
# liability. 
# 
# In this respect, the user's attention is drawn to the risks associated
# with loading,  using,  modifying and/or developing or reproducing the
# software by the user in light of its specific status of free software,
# that may mean  that it is complicated to manipulate,  and  that  also
# therefore means  that it is reserved for developers  and  experienced
# professionals having in-depth computer knowledge. Users are therefore
# encouraged to load and test the software's suitability as regards their
# requirements in conditions enabling the security of their systems and/or 
# data to be ensured and,  more generally, to use and operate it in the 
# same conditions as regards security. 
# 
# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.
# ----------------------------------------------------------------------------------
# 
'''
Reset mechanisms
'''

__all__ = ['Reset', 'VariableReset', 'Refractoriness', 'NoReset', 'FunReset',
         'CustomRefractoriness', 'SimpleCustomRefractoriness', 'StringReset',
         'select_reset']

from numpy import where, zeros
from units import *
from clock import *
import inspect
import re
import numpy
from inspection import *
from utils.documentation import flattened_docstring
from globalprefs import *
from log import *
CReset = PythonReset = None

def select_reset(expr, eqs, level=0):
    '''
    Automatically selects the appropriate Reset object from a string.
    
    Matches the following patterns if expr is a one liner:
    
    var_name = const : Reset
    var_name = var_name : VariableReset
    others : StringReset
    '''
    # plan:
    # - strip it and see if it is one line, if not select StringReset
    # - see if it matches A = B, if not select StringReset
    # - check if A, B both match diffeq variable names, and if so
    #   select VariableReset
    # - check that A is a variable name, if not select StringReset
    # - extract all the identifiers from B, and if none of them are
    #   callable, assume it is a constant, try to eval it and then use
    #   Reset. If not, or if eval fails, use StringReset
    # This misses the case of e.g. V=10*mV*exp(1) because exp will be
    # callable, but in general a callable means that it could be
    # non-constant.
    global CReset, PythonReset
    use_codegen = get_global_preference('usecodegen') and get_global_preference('usecodegenreset')
    use_weave = get_global_preference('useweave') and get_global_preference('usecodegenweave')
    if use_codegen:
        if CReset is None:
            from experimental.codegen.reset import CReset, PythonReset
        if use_weave:
            log_warn('brian.reset', 'Using codegen CReset')
            return CReset(expr, level=level + 1)
        else:
            log_warn('brian.reset', 'Using codegen PythonReset')
            return PythonReset(expr, level=level + 1)
    expr = expr.strip()
    if '\n' in expr:
        return StringReset(expr, level=level + 1)
    eqs.prepare()
    ns = namespace(expr, level=level + 1)
    s = re.search(r'\s*(\w+)\s*=(.+)', expr)
    if not s:
        return StringReset(expr, level=level + 1)
    A = s.group(1)
    B = s.group(2).strip()
    if A not in eqs._diffeq_names:
        return StringReset(expr, level=level + 1)
    if B in eqs._diffeq_names:
        return VariableReset(B, A)
    vars = get_identifiers(B)
    all_vars = eqs._eq_names + eqs._diffeq_names + eqs._alias.keys() + ['t']
    for v in vars:
        if v not in ns or v in all_vars or callable(ns[v]):
            return StringReset(expr, level=level + 1)
    try:
        val = eval(B, ns)
    except:
        return StringReset(expr, level=level + 1)
    return Reset(val, A)


class Reset(object):
    '''
    Resets specified state variable to a fixed value
    
    **Initialise as:** ::
    
        R = Reset([resetvalue=0*mvolt[, state=0]])
        
    with arguments:
    
    ``resetvalue``
        The value to reset to.
    ``state``
        The name or number of the state variable to reset.

    This will reset all of the neurons that have just spiked. The
    given state variable of the neuron group will be set to value
    ``resetvalue``.
    '''
    def __init__(self, resetvalue=0 * mvolt, state=0):
        self.resetvalue = resetvalue
        self.state = state
        self.statevectors = {}

    def __call__(self, P):
        '''
        Clamps membrane potential at reset value.
        '''
        V = self.statevectors.get(id(P), None)
        if V is None:
            V = P.state_(self.state)
            self.statevectors[id(P)] = V
        V[P.LS.lastspikes()] = self.resetvalue

    def __repr__(self):
        return 'Reset ' + str(self.resetvalue)


class StringReset(Reset):
    '''
    Reset defined by a string
    
    Initialised with arguments:
    
    ``expr``
        The string expression used to reset. This can include 
        multiple lines or statements separated by a semicolon.
        For example, ``'V=-70*mV'`` or ``'V=-70*mV; Vt+=10*mV'``.
        Some standard functions are provided, see below.
    ``level``
        How many levels up in the calling sequence to look for
        names in the namespace. Usually 0 for user code.
    
    Standard functions for expressions:
    
    ``rand()``
        A uniform random number between 0 and 1.
    ``randn()``
        A Gaussian random number with mean 0 and standard deviation 1.
    
    For example, these could be used to implement an adaptive
    model with random reset noise with the following string::
    
        E -= 1*mV
        V = Vr+rand()*5*mV
    '''
    def __init__(self, expr, level=0):
        expr = flattened_docstring(expr)
        self._namespace, unknowns = namespace(expr, level=level + 1, return_unknowns=True)
        self._prepared = False
        self._expr = expr
        class Replacer(object):
            def __init__(self, func, n):
                self.n = n
                self.func = func
            def __call__(self):
                return self.func(self.n)
        self._Replacer = Replacer

    def __call__(self, P):
        if not self._prepared:
            unknowns = [var for var in P.var_index if isinstance(var, str)]
            expr = self._expr
            for var in unknowns:
                expr = re.sub("\\b" + var + "\\b", var + '[_spikes_]', expr)
            self._code = compile(expr, "StringReset", "exec")
            self._vars = unknowns
            self._prepared = True
        spikes = P.LS.lastspikes()
        self._namespace['_spikes_'] = spikes
        self._namespace['rand'] = self._Replacer(numpy.random.rand, len(spikes))
        self._namespace['randn'] = self._Replacer(numpy.random.randn, len(spikes))
        for var in self._vars:
            self._namespace[var] = P.state(var)
        exec self._code in self._namespace

    def __repr__(self):
        return "String reset"


class VariableReset(Reset):
    '''
    Resets specified state variable to the value of another state variable
    
    Initialised with arguments:
    
    ``resetvaluestate``
        The state variable which contains the value to reset to.
    ``state``
        The name or number of the state variable to reset.

    This will reset all of the neurons that have just spiked. The
    given state variable of the neuron group will be set to
    the value of the state variable ``resetvaluestate``.
    '''
    def __init__(self, resetvaluestate=1, state=0):
        self.resetvaluestate = resetvaluestate
        self.state = state
        self.resetstatevectors = {}
        self.statevectors = {}

    def __call__(self, P):
        '''
        Clamps membrane potential at reset value.
        '''
        V = self.statevectors.get(id(P), None)
        if V is None:
            V = P.state_(self.state)
            self.statevectors[id(P)] = V
        Vr = self.resetstatevectors.get(id(P), None)
        if Vr is None:
            Vr = P.state_(self.resetvaluestate)
            self.resetstatevectors[id(P)] = Vr
        lastspikes = P.LS.lastspikes()
        V[lastspikes] = Vr[lastspikes]

    def __repr__(self):
        return 'VariableReset(' + str(self.resetvaluestate) + ', ' + str(self.state) + ')'


class FunReset(Reset):
    '''
    A reset with a user-defined function.
    
    **Initialised as:** ::
    
        FunReset(resetfun)
    
    with argument:
    
    ``resetfun``
        A function ``f(G,spikes)`` where ``G`` is the
        :class:`NeuronGroup` and ``spikes`` is an array of
        the indexes of the neurons to be reset.
    '''
    def __init__(self, resetfun):
        self.resetfun = resetfun

    def __call__(self, P):
        self.resetfun(P, P.LS.lastspikes())


class Refractoriness(Reset):
    '''
    Holds the state variable at the reset value for a fixed time after a spike.

    Initialised with arguments:
    
    ``resetvalue``
        The value to reset and hold to.
    ``period``
        The length of time to hold at the reset value. If using variable
        refractoriness, this is the maximum period.
    ``state``
        The name or number of the state variable to reset and hold.
    '''
    @check_units(period=second)
    def __init__(self, resetvalue=0 * mvolt, period=5 * msecond, state=0):
        self.period = period
        self.resetvalue = resetvalue
        self.state = state
        self._periods = {} # a dictionary mapping group IDs to periods
        self.statevectors = {}

    def __call__(self, P):
        '''
        Clamps state variable at reset value.
        '''
        # if we haven't computed the integer period for this group yet.
        # do so now
        if id(P) in self._periods:
            period = self._periods[id(P)]
        else:
            period = int(self.period / P.clock.dt) + 1
            self._periods[id(P)] = period
        V = self.statevectors.get(id(P), None)
        if V is None:
            V = P.state_(self.state)
            self.statevectors[id(P)] = V
        neuronindices = P.LS[0:period]
        if P._variable_refractory_time:
            neuronindices = neuronindices[P._next_allowed_spiketime[neuronindices] > (P.clock._t - P.clock._dt * 0.25)]
        V[neuronindices] = self.resetvalue

    def __repr__(self):
        return 'Refractory period, ' + str(self.period)


class SimpleCustomRefractoriness(Refractoriness):
    '''
    Holds the state variable at the custom reset value for a fixed time after a spike.
    
    **Initialised as:** ::
    
        SimpleCustomRefractoriness(resetfunc[,period=5*ms[,state=0]])
    
    with arguments:
    
    ``resetfun``
        The custom reset function ``resetfun(P, spikes)`` for ``P`` a
        :class:`NeuronGroup` and ``spikes`` a list of neurons that
        fired spikes.
    ``period``
        The length of time to hold at the reset value.
    ``state``
        The name or number of the state variable to reset and hold,
        it is your responsibility to check that this corresponds to
        the custom reset function.
    
    The assumption is that ``resetfun(P, spikes)`` will reset the state
    variable ``state`` on the group ``P`` for the spikes with indices
    ``spikes``. The values assigned by the custom reset function are
    stored by this object, and they are clamped at these values for
    ``period``. This object does not introduce refractoriness for more
    than the one specified variable ``state`` or for spike indices
    other than those in the variable ``spikes`` passed to the custom
    reset function.
    '''

    @check_units(period=second)
    def __init__(self, resetfun, period=5 * msecond, state=0):
        self.period = period
        self.resetfun = resetfun
        self.state = state
        self._periods = {} # a dictionary mapping group IDs to periods
        self.statevectors = {}
        self.lastresetvalues = {}

    def __call__(self, P):
        '''
        Clamps state variable at reset value.
        '''
        # if we haven't computed the integer period for this group yet.
        # do so now
        if id(P) in self._periods:
            period = self._periods[id(P)]
        else:
            period = int(self.period / P.clock.dt) + 1
            self._periods[id(P)] = period
        V = self.statevectors.get(id(P), None)
        if V is None:
            V = P.state_(self.state)
            self.statevectors[id(P)] = V
        LRV = self.lastresetvalues.get(id(P), None)
        if LRV is None:
            LRV = zeros(len(V))
            self.lastresetvalues[id(P)] = LRV
        lastspikes = P.LS.lastspikes()
        self.resetfun(P, lastspikes)             # call custom reset function 
        LRV[lastspikes] = V[lastspikes]         # store a copy of the custom resetted values
        clampedindices = P.LS[0:period]
        V[clampedindices] = LRV[clampedindices] # clamp at custom resetted values

    def __repr__(self):
        return 'Custom refractory period, ' + str(self.period)


class CustomRefractoriness(Refractoriness):
    '''
    Holds the state variable at the custom reset value for a fixed time after a spike.
    
    **Initialised as:** ::
    
        CustomRefractoriness(resetfunc[,period=5*ms[,refracfunc=resetfunc]])
    
    with arguments:
    
    ``resetfunc``
        The custom reset function ``resetfunc(P, spikes)`` for ``P`` a
        :class:`NeuronGroup` and ``spikes`` a list of neurons that
        fired spikes.
    ``refracfunc``
        The custom refractoriness function ``refracfunc(P, indices)`` for ``P`` a
        :class:`NeuronGroup` and ``indices`` a list of neurons that are in
        their refractory periods. In some cases, you can choose not to specify this,
        and it will use the reset function.
    ``period``
        The length of time to hold at the reset value.    
    '''

    @check_units(period=second)
    def __init__(self, resetfun, period=5 * msecond, refracfunc=None):
        self.period = period
        self.resetfun = resetfun
        if refracfunc is None:
            refracfunc = resetfun
        self.refracfunc = refracfunc
        self._periods = {} # a dictionary mapping group IDs to periods

    def __call__(self, P):
        '''
        Clamps state variable at reset value.
        '''
        # if we haven't computed the integer period for this group yet.
        # do so now
        if id(P) in self._periods:
            period = self._periods[id(P)]
        else:
            period = int(self.period / P.clock.dt) + 1
            self._periods[id(P)] = period
        lastspikes = P.LS.lastspikes()
        self.resetfun(P, lastspikes)             # call custom reset function
        clampedindices = P.LS[0:period]
        self.refracfunc(P, clampedindices)

    def __repr__(self):
        return 'Custom refractory period, ' + str(self.period)


class NoReset(Reset):
    '''
    Absence of reset mechanism.
    
    **Initialised as:** ::
    
        NoReset()
    '''
    def __init__(self):
        pass

    def __call__(self, P):
        pass

    def __repr__(self):
        return 'No reset'



from brian import *
#This code leaks memory if brianhears_usegpu is switched on
set_global_preferences(brianhears_usegpu=True)
from brian.hears import *

sound = silence(1 * second)
gfb = Gammatone(sound, 100)
n = 1000
for i in xrange(n):
    if (i + 1) % 10 == 0:
        print '%d/%d' % (i+1, n)
    gfb.process()
    


import pycuda.autoinit as autoinit
import pycuda.compiler as drv # .driver before pycuda 0.93
from pycuda.gpuarray import GPUArray
from pycuda import gpuarray
import numpy, pylab, time

N = 10000
duration = int(0.1 * 10000)
record = False
showfinal = True
doget = True
dosync = False
block = (512, 1, 1)
grid = (int(N / 512) + 1, 1)

#Device attributes
#MAX_THREADS_PER_BLOCK : 512
#MAX_BLOCK_DIM_X : 512
#MAX_BLOCK_DIM_Y : 512
#MAX_BLOCK_DIM_Z : 64
#MAX_GRID_DIM_X : 65535
#MAX_GRID_DIM_Y : 65535
#MAX_GRID_DIM_Z : 1
#MAX_SHARED_MEMORY_PER_BLOCK : 16384
#TOTAL_CONSTANT_MEMORY : 65536
#WARP_SIZE : 32
#MAX_PITCH : 262144
#MAX_REGISTERS_PER_BLOCK : 16384
#CLOCK_RATE : 1296000
#TEXTURE_ALIGNMENT : 256
#GPU_OVERLAP : 1
#MULTIPROCESSOR_COUNT : 30

mod = drv.SourceModule("""
__global__ void stateupdate(double *v, double *w)
{
  //const int i = threadIdx.x;
  int i = blockIdx.x * blockDim.x + threadIdx.x;
  double v__tmp = w[i]*w[i]/(100*0.001);
  double w__tmp = -v[i]/(100*0.001);
  v[i] += 0.0001*v__tmp;
  w[i] += 0.0001*w__tmp;
}
""")

stateupdate = mod.get_function("stateupdate")

v = gpuarray.to_gpu(numpy.ones(N))
w = gpuarray.to_gpu(numpy.zeros(N))

#ov = numpy.zeros(N)
ov = drv.pagelocked_zeros((N,), dtype=float)

recv = []
start = time.time()
for _ in xrange(duration):
    stateupdate(v, w, block=block, grid=grid)
    if record:
        recv.append(v.get())
    elif doget:
        v.get(ov)
    elif dosync:
        autoinit.context.synchronize()
print 'GPU code:', time.time() - start

if record:
    x = numpy.array(recv)
    pylab.plot(x[:, 0])
if showfinal:
    pylab.figure()
    pylab.plot(v.get())
if record or showfinal:
    pylab.show()

'''
Some results:

dV/dt = W*W/(100*ms) : 1
dW/dt = -V/(100*ms) : 1

10k neurons for 1000 time steps
Python: 0.5s
C++: 0.4s
GPUgetpagelocked: 0.1s 
GPUsync: 0.047s

100k neurons
Python: 9.3s
C++: 2.9s
GPUgetpagelocked: 0.45s
GPUsync: 0.16s

1M neurons
Python: 111s
C++: 27.6s
GPUgetpagelocked: 4.1s
GPUsync: 1.2s
'''



import numpy
import time
import random
from itertools import repeat

# units
second = 1.
ms = 0.001
mV = 0.001

# Parameters
N = 4000
Ne = int(N * 0.8)
Ni = N - Ne
Nplot = 4
dt = 0.1 * ms
T = 1 * second
numsteps = int(T / dt)
p = 0.02
Vr = 0 * mV
Vt = 10 * mV
we = 6. / 10. # excitatory synaptic weight (voltage)
wi = 67. / 10. # inhibitory synaptic weight
refrac = 5 * ms

# State variable S=[v;ge;gi] and variable used in Euler step
# dS=[v';ge';gi'] used in the main loop below
S = numpy.zeros((3, N))
dS = numpy.zeros((3, N))
v, ge, gi = S
v__tmp, ge__tmp, gi__tmp = dS

# Last spike times, stores the most recent time that a neuron
# has spiked, which is used for refractory periods
LS = numpy.zeros(N)
LS[:] = -2 * refrac

# Initialisation of state variables
S[0, :] = (numpy.random.randn(N) * 5 - 5) * mV
S[1, :] = numpy.random.randn(N) * 1.5 + 4
S[2, :] = numpy.random.randn(N) * 12 + 20

# Weight matrix
# Generate random connectivity matrix (note: no weights)
W = []
for _ in xrange(N):
    k = numpy.random.binomial(N, p, 1)[0]
    a = random.sample(xrange(N), k)
    a.sort()
    a = numpy.array(a)
    W.append(a)

Vrec = [[] for _ in range(Nplot)]
spikesrec = []

t = 0.
Nspikes = 0
start = time.time()
for _ in xrange(numsteps):
    # EULER UPDATE CODE, this is the update code generated by Brian for the equations:
    #dv/dt = (-v+ge*(Ee-v)+gi*(Ei-v))*(1./taum) : volt
    #dge/dt = -ge*(1./taue) : 1
    #dgi/dt = -gi*(1./taui) : 1 
    v__tmp[:] = (-v + ge * (0.06 - v) + gi * (-0.02 - v)) * (1. / 0.02)
    ge__tmp[:] = -ge * (1. / 0.005)
    gi__tmp[:] = -gi * (1. / 0.01)
    S += dt * dS

    # spike propagation, essentially the same as CUBA except I've used the names v,ge,gi
    # here instead of S[0,:] etc.
    spikes = (v > Vt).nonzero()[0]
    for i in spikes:
        if i < Ne:
            ge[W[i]] += we
        else:
            gi[W[i]] += wi

    # refractory period, if the last spike of neuron i was within refrac of the current time,
    # fix the value to Vr - note that this is how Brian used to do this, but now we have a
    # slightly different mechanism, but this is very close.
    LS[spikes] = t
    v[LS > t - refrac] = Vr

    # Monitoring, actually takes about 1/3 of the time of the simulation when I ran it,
    # for speed comparisons this bit should be commented out
    Nspikes += len(spikes)
    spikesrec += zip(spikes, repeat(t))
    for i in range(Nplot):
        Vrec[i].append(S[0, i])

    t += dt

print 'Time taken', time.time() - start
print 'Spikes', Nspikes

try:
    import pylab
    if len(spikesrec):
        sn, st = zip(*spikesrec)
        pylab.subplot(121)
        pylab.plot(st, sn, '.')
        pylab.subplot(122)
        for i in range(Nplot):
            pylab.plot(Vrec[i])
        pylab.show()
except ImportError:
    print 'Cannot plot, no pylab'



import re
from rewriting import *
from ...inspection import namespace
from ...optimiser import freeze
from ...reset import Reset
from ...equations import Equations
from ...globalprefs import get_global_preference
from ...log import log_warn
from expressions import *
from scipy import weave
from c_support_code import *

__all__ = ['generate_c_reset', 'generate_python_reset',
           'CReset', 'PythonReset']

def generate_c_reset(eqs, inputcode, vartype='double', level=0, ns=None):
    if ns is None:
        ns, unknowns = namespace(inputcode, level=level + 1, return_unknowns=True)
    all_variables = eqs._eq_names + eqs._diffeq_names + eqs._alias.keys() + ['t']
    code = ''
    for j, name in enumerate(eqs._diffeq_names):
        code += vartype + ' *' + name + '__Sbase = _S+' + str(j) + '*_num_neurons;\n'
    code += 'for(int _i=0;_i<_nspikes;_i++){\n'
    code += '    long _j = _spikes[_i];\n'
    for j, name in enumerate(eqs._diffeq_names):
        code += '    ' + vartype + ' &' + name + ' = ' + name + '__Sbase[_j];\n'
    for line in inputcode.split('\n'):
        line = line.strip()
        if line:
            line = freeze(line, all_variables, ns)
            line = c_single_statement(line)
            code += '    ' + line + '\n'
    code += '}\n'
    return code

def generate_python_reset(eqs, inputcode, level=0, ns=None):
    if ns is None:
        ns, unknowns = namespace(inputcode, level=level + 1, return_unknowns=True)
    all_variables = eqs._eq_names + eqs._diffeq_names + eqs._alias.keys() + ['t', '_spikes']
    code = ''
    for var in eqs._diffeq_names:
        inputcode = re.sub("\\b" + var + "\\b", var + '[_spikes]', inputcode)
    for line in inputcode.split('\n'):
        line = line.strip()
        if line:
            line = freeze(line, all_variables, ns)
            line = python_single_statement(line)
            code += line + '\n'
    return code


class PythonReset(Reset):
    def __init__(self, inputcode, level=0):
        self._ns, unknowns = namespace(inputcode, level=level + 1, return_unknowns=True)
        self._inputcode = inputcode
        self._prepared = False

    def __call__(self, P):
        ns = self._ns
        if not self._prepared:
            vars = [var for var in P.var_index if isinstance(var, str)]
            eqs = P._eqs
            outputcode = generate_python_reset(eqs, self._inputcode, ns=ns)
            self._compiled_code = compile(outputcode, "PythonReset", "exec")
            for var in vars:
                ns[var] = P.state(var)
            self._prepared = True
        cc = self._compiled_code
        spikes = P.LS.lastspikes()
        ns['_spikes'] = spikes
        exec cc in ns


class CReset(Reset):
    def __init__(self, inputcode, level=0):
        self._ns, unknowns = namespace(inputcode, level=level + 1, return_unknowns=True)
        self._inputcode = inputcode
        self._prepared = False
        self._weave_compiler = get_global_preference('weavecompiler')
        self._extra_compile_args = ['-O3']
        if self._weave_compiler == 'gcc':
            self._extra_compile_args += get_global_preference('gcc_options') # ['-march=native', '-ffast-math']

    def __call__(self, P):
        if not self._prepared:
            vars = [var for var in P.var_index if isinstance(var, str)]
            eqs = P._eqs
            self._outputcode = generate_c_reset(eqs, self._inputcode, ns=self._ns)
        _spikes = P.LS.lastspikes()
        dt = P.clock._dt
        t = P.clock._t
        _nspikes = len(_spikes)
        _S = P._S
        _num_neurons = len(P)
        try:
            weave.inline(self._outputcode, ['_S', '_nspikes', 'dt', 't', '_spikes', '_num_neurons'],
                         c_support_code=c_support_code,
                         compiler=self._weave_compiler,
                         extra_compile_args=self._extra_compile_args)
        except:
            log_warn('brian.experimental.codegen.reset',
                     'C compilation failed, falling back on Python.')
            self.__class__ = PythonReset
            self._prepared = False
            return self.__call__(P)



'''
Example of online computation using :meth:`~brian.hears.Filterbank.process`.
Plots the RMS value of each channel output by a gammatone filterbank.
'''
from brian import *
from brian.hears import *

sound1 = tone(1*kHz, .1*second)
sound2 = whitenoise(.1*second)

sound = sound1+sound2
sound = sound.ramp()

sound.level = 60*dB

cf = erbspace(20*Hz, 20*kHz, 3000)
fb = Gammatone(sound, cf)

def sum_of_squares(input, running):
    return running+sum(input**2, axis=0)

rms = sqrt(fb.process(sum_of_squares)/sound.nsamples)

sound_rms = sqrt(mean(sound**2))

axhline(sound_rms, ls='--')
plot(cf, rms)
xlabel('Frequency (Hz)')
ylabel('RMS')
show()



from numpy import *
from numpy.random import *
import time

N = 150

bx = randn(N, N)
by = randn(N, N)

t = time.time()
for i in xrange(1000):
    x = array(bx, copy=True)
    y = array(by, copy=True)
baset = time.time() - t

t = time.time()
for i in xrange(1000):
    x = matrix(bx, copy=True)
    y = matrix(by, copy=True)
basebt = time.time() - t

t = time.time()
for i in xrange(1000):
    x = array(bx, copy=True)
    y = array(by, copy=True)
    z = dot(x, y)
s = time.time() - t
print s, s - baset

t = time.time()
for i in xrange(1000):
    x = matrix(bx, copy=True)
    y = matrix(by, copy=True)
    z = x * y
s = time.time() - t
print s, s - basebt

t = time.time()
for i in xrange(1000):
    x = matrix(bx, copy=True)
    y = matrix(by, copy=True)
    x *= y
s = time.time() - t
print s, s - basebt



'''
A very simple example Brian script to show how to implement
a leaky integrate and fire model. In this example, we also
drive the single leaky integrate and fire neuron with
regularly spaced spikes from the :class:`SpikeGeneratorGroup`.
'''
from brian import *

tau = 10 * ms
Vr = -70 * mV
Vt = -55 * mV

G = NeuronGroup(1, model='dV/dt = -(V-Vr)/tau : volt', threshold=Vt, reset=Vr)

spikes = linspace(10 * ms, 100 * ms, 25)
input = MultipleSpikeGeneratorGroup([spikes])

C = Connection(input, G)
C[0, 0] = 5 * mV

M = StateMonitor(G, 'V', record=True)

G.V = Vr
run(100 * ms)
plot(M.times / ms, M[0] / mV)
show()



from brian.tools.statistics import *
from brian.units import *
from brian.stdunits import *
from numpy import *
from numpy.random import *
from nose.tools import *

def test_group_correlations():
    rate = 10.0
    n = 100
    poisson = cumsum(exponential(1 / rate, n))
    spikes = [(0, t * second) for t in poisson]
    spikes.extend([(0, t * second + 1 * ms) for t in poisson])
    spikes = sort_spikes(spikes)
    S, tauc = group_correlations(spikes, delta=3 * ms)
    assert abs(tauc[0] - .001) < .0005

if __name__ == '__main__':
    test_group_correlations()



"""
This is a modified version of source code from the Accerciser project
(http://live.gnome.org/accerciser).

Backend to the console plugin.

@author: Eitan Isaacson
@organization: IBM Corporation
@copyright: Copyright (c) 2007 IBM Corporation
@license: BSD

All rights reserved. This program and the accompanying materials are made 
available under the terms of the BSD which accompanies this distribution, and 
is available at U{http://www.opensource.org/licenses/bsd-license.php}
"""

import re
import sys
import os
from StringIO import StringIO
import Tkinter
import IPython


class IterableIPShell:
  def __init__(self, argv=None, user_ns=None, user_global_ns=None,
               cin=None, cout=None, cerr=None, input_func=None):
    if input_func:
      IPython.iplib.raw_input_original = input_func
    if cin:
      IPython.Shell.Term.cin = cin
    if cout:
      IPython.Shell.Term.cout = cout
    if cerr:
      IPython.Shell.Term.cerr = cerr

    if argv is None:
      argv = []

    # This is to get rid of the blockage that occurs during 
    # IPython.Shell.InteractiveShell.user_setup()
    IPython.iplib.raw_input = lambda x: None

    self.term = IPython.genutils.IOTerm(cin=cin, cout=cout, cerr=cerr)
    os.environ['TERM'] = 'dumb'
    excepthook = sys.excepthook
    self.IP = IPython.Shell.make_IPython(argv, user_ns=user_ns,
                                         user_global_ns=user_global_ns,
                                         embedded=True,
                                         shell_class=IPython.Shell.InteractiveShell)
    self.IP.system = lambda cmd: self.shell(self.IP.var_expand(cmd),
                                            header='IPython system call: ',
                                            verbose=self.IP.rc.system_verbose)
    sys.excepthook = excepthook
    self.iter_more = 0
    self.history_level = 0
    self.complete_sep = re.compile('[\s\{\}\[\]\(\)]')

  def execute(self):
    self.history_level = 0
    orig_stdout = sys.stdout
    sys.stdout = IPython.Shell.Term.cout
    try:
      line = self.IP.raw_input(None, self.iter_more)
      if self.IP.autoindent:
        self.IP.readline_startup_hook(None)
    except KeyboardInterrupt:
      self.IP.write('\nKeyboardInterrupt\n')
      self.IP.resetbuffer()
      # keep cache in sync with the prompt counter:
      self.IP.outputcache.prompt_count -= 1

      if self.IP.autoindent:
        self.IP.indent_current_nsp = 0
      self.iter_more = 0
    except:
      self.IP.showtraceback()
    else:
      self.iter_more = self.IP.push(line)
      if (self.IP.SyntaxTB.last_syntax_error and
          self.IP.rc.autoedit_syntax):
        self.IP.edit_syntax_error()
    if self.iter_more:
      self.prompt = str(self.IP.outputcache.prompt2).strip()
      if self.IP.autoindent:
        self.IP.readline_startup_hook(self.IP.pre_readline)
    else:
      self.prompt = str(self.IP.outputcache.prompt1).strip()
    sys.stdout = orig_stdout

  def historyBack(self):
    self.history_level -= 1
    return self._getHistory()

  def historyForward(self):
    self.history_level += 1
    return self._getHistory()

  def _getHistory(self):
    try:
      rv = self.IP.user_ns['In'][self.history_level].strip('\n')
    except IndexError:
      self.history_level = 0
      rv = ''
    return rv

  def updateNamespace(self, ns_dict):
    self.IP.user_ns.update(ns_dict)

  def complete(self, line):
    split_line = self.complete_sep.split(line)
    possibilities = self.IP.complete(split_line[-1])
    if possibilities:
      common_prefix = reduce(self._commonPrefix, possibilities)
      completed = line[:-len(split_line[-1])] + common_prefix
    else:
      completed = line
    return completed, possibilities

  def _commonPrefix(self, str1, str2):
    for i in range(len(str1)):
      if not str2.startswith(str1[:i + 1]):
        return str1[:i]
    return str1

  def shell(self, cmd, verbose=0, debug=0, header=''):
    stat = 0
    if verbose or debug: print header + cmd
    # flush stdout so we don't mangle python's buffering
    if not debug:
      input, output = os.popen4(cmd)
      print output.read()
      output.close()
      input.close()


ansi_colors = {'0;30': 'Black',
                '0;31': 'Red',
                '0;32': 'Green',
                '0;33': 'Brown',
                '0;34': 'Blue',
                '0;35': 'Purple',
                '0;36': 'Cyan',
                '0;37': 'LightGray',
                '1;30': 'DarkGray',
                '1;31': 'DarkRed',
                '1;32': 'SeaGreen',
                '1;33': 'Yellow',
                '1;34': 'LightBlue',
                '1;35': 'MediumPurple',
                '1;36': 'LightCyan',
                '1;37': 'White'}


class TkConsoleView(Tkinter.Text):
  def __init__(self, root):
    Tkinter.Text.__init__(self, root)

    # As the stdout,stderr etc. get fiddled about with we need to put any
    # debug output into a file
    self.debug = 0
    if self.debug:
            self.o = open('debug.out', 'w')

    # Keeps track of where the insert cursor should be on the entry line
    self.mark = 'scroll_mark'
    self.mark_set(self.mark, Tkinter.END)
    self.mark_gravity(self.mark, Tkinter.RIGHT)

    # Set the tags for colouring the text
    for code in ansi_colors:
      self.tag_config(code,
                      foreground=ansi_colors[code])

    self.tag_config('notouch') # Tag for indicating what areas of the widget aren't editable


    # colour_pat matches the colour tags and places these in a group
    # match character with hex value 01 (start of heading?) zero or more times, followed by
    # the hex character 1b (escape)  then "[" and group ...things.. followed by m (?) and then
    # hex character 02 (start of text) zero or more times
    self.color_pat = re.compile('\x01?\x1b\[(.*?)m\x02?')

    self.line_start = 'line_start' # Tracks start of user input on the line (excluding prompt)
    self.mark_set(self.line_start, Tkinter.INSERT)
    self.mark_gravity(self.line_start, Tkinter.LEFT)

    self._setBindings()

  def write(self, text, editable=False):

    segments = self.color_pat.split(text)
    # First is blank line
    segment = segments.pop(0)

    # Keep track of where we started entering text so we can set as non-editable
    self.start_mark = 'start_mark'
    self.mark_set(self.start_mark, Tkinter.INSERT)
    self.mark_gravity(self.start_mark, Tkinter.LEFT)

    self.insert(Tkinter.END, segment)

    if segments:
      # Just return the colour tags
      ansi_tags = self.color_pat.findall(text)

      for tag in ansi_tags:
        i = segments.index(tag)
        self.insert(Tkinter.END, segments[i + 1], tag)
        segments.pop(i)

    if not editable:
      if self.debug:
          print "adding notouch between %s : %s" % (self.index(self.start_mark), \
                                                     self.index(Tkinter.INSERT))

      self.tag_add('notouch', self.start_mark, "%s-1c" % Tkinter.INSERT)

    self.mark_unset(self.start_mark)
    #jmht self.scroll_mark_onscreen(self.mark)


  def showBanner(self, banner):
    """Print the supplied banner on starting the shell"""
    self.write(banner)

  def showPrompt(self, prompt):
    self.write(prompt)
    self.mark_set(self.line_start, Tkinter.INSERT)
    self.see(Tkinter.INSERT) #Make sure we can always see the prompt


  def changeLine(self, text):
    self.delete(self.line_start, "%s lineend" % self.line_start)
    self.write(text, True)

  def getCurrentLine(self):

    rv = self.get(self.line_start, Tkinter.END)

    if self.debug:
        print >> self.o, "getCurrentline: %s" % rv
        print >> self.o, "INSERT: %s" % Tkinter.END
        print >> self.o, "END: %s" % Tkinter.INSERT
        print >> self.o, "line_start: %s" % self.index(self.line_start)

    return rv

  def showReturned(self, text):
    self.tag_add('notouch', self.line_start, "%s lineend" % self.line_start)
    self.write('\n' + text)
    if text:
      self.write('\n')
    self.showPrompt(self.prompt)
    #self.mark_set(self.line_start,Tkinter.END) #jmht don't need this as showprompt sets mark

  def _setBindings(self):
    """ Bind the keys we require.
        REM: if a bound function returns "break" then no other bindings are called
        If it returns None, then the other default bindings are called.
    """
    self.bind("<Key>", self.processKeyPress)
    self.bind("<Return>", self.processEnterPress)
    self.bind("<Up>", self.processUpPress)
    self.bind("<Down>", self.processDownPress)
    self.bind("<Tab>", self.processTabPress)
    self.bind("<BackSpace>", self.processBackSpacePress)

  def isEditable(self):
    """ Scan the notouch tag range in pairs and see if the INSERT index falls
        between any of them.
    """
    ranges = self.tag_ranges('notouch')
    first = None
    for idx in ranges:
        if not first:
            first = idx
            continue
        else:

            if self.debug:
                print "Comparing %s between %s : %s " % (self.index(Tkinter.INSERT), first, idx)

            if self.compare(Tkinter.INSERT, '>=', first) and \
                   self.compare(Tkinter.INSERT, '<=', idx):
                return False
            first = None
    return True

  def processKeyPress(self, event):

    if self.debug:
        print >> self.o, "processKeyPress got key: %s" % event.char
        print >> self.o, "processKeyPress INSERT: %s" % self.index(Tkinter.INSERT)
        print >> self.o, "processKeyPress END: %s" % self.index(Tkinter.END)

    if not self.isEditable():
            # Move cursor mark to start of line
            self.mark_set(Tkinter.INSERT, self.mark)

    # Make sure line_start follows inserted text
    self.mark_set(self.mark, "%s+1c" % Tkinter.INSERT)


  def processBackSpacePress(self, event):
    if not self.isEditable():
            return "break"

  def processEnterPress(self, event):
    self._processLine()
    return "break" # Need break to stop the other bindings being called

  def processUpPress(self, event):
    self.changeLine(self.historyBack())
    return "break"

  def processDownPress(self, event):
    self.changeLine(self.historyForward())
    return "break"

  def processTabPress(self, event):
    if not self.getCurrentLine().strip():
            return
    completed, possibilities = self.complete(self.getCurrentLine())
    if len(possibilities) > 1:
            slice = self.getCurrentLine()
            self.write('\n')
            for symbol in possibilities:
                    self.write(symbol + '\n')
            self.showPrompt(self.prompt)
    self.changeLine(completed or slice)
    return "break"


class IPythonView(TkConsoleView, IterableIPShell):
  def __init__(self, root, banner=None):
    TkConsoleView.__init__(self, root)
    self.cout = StringIO()
    IterableIPShell.__init__(self, cout=self.cout, cerr=self.cout,
                             input_func=self.raw_input)

    if banner:
      self.showBanner(banner)
    self.execute()
    self.cout.truncate(0)
    self.showPrompt(self.prompt)
    self.interrupt = False

  def raw_input(self, prompt=''):
    if self.interrupt:
      self.interrupt = False
      raise KeyboardInterrupt
    return self.getCurrentLine()

  def _processLine(self):
    self.history_pos = 0
    self.execute()
    rv = self.cout.getvalue()
    if self.debug:
        print >> self.o, "_processLine got rv: %s" % rv
    if rv: rv = rv.strip('\n')
    self.showReturned(rv)
    self.cout.truncate(0)

if __name__ == "__main__":
    root = Tkinter.Tk()
    s = IPythonView(root)
    s.pack()
    root.mainloop()



'''
GPU Buffering class

See docstring for GPUBufferedArray for details.
'''
from brian import *
import pycuda
import pycuda.gpuarray
import numpy
from numpy import intp

__all__ = ['SynchronisationError', 'GPUBufferedArray']

if __name__ == '__main__':
    DEBUG_BUFFER_CACHE = True
else:
    DEBUG_BUFFER_CACHE = False

numpy_inplace_methods = set([
     '__iadd__',
     '__iand__',
     '__idiv__',
     '__ifloordiv__',
     '__ilshift__',
     '__imod__',
     '__imul__',
     '__ior__',
     '__ipow__',
     '__irshift__',
     '__isub__',
     '__itruediv__',
     '__ixor__',
     '__setitem__',
     '__setslice__',
     'byteswap',
     'fill',
     'put',
     'sort',
     ])

numpy_access_methods = set([
     '__abs__',
     '__add__',
     '__and__',
     '__array__',
     '__contains__',
     '__copy__',
     '__deepcopy__',
     '__delattr__',
     '__delitem__',
     '__delslice__',
     '__div__',
     '__divmod__',
     '__eq__',
     '__float__',
     '__floordiv__',
     '__ge__',
     '__getitem__',
     '__getslice__',
     '__gt__',
     '__hash__',
     '__hex__',
     '__iadd__',
     '__iand__',
     '__idiv__',
     '__ifloordiv__',
     '__ilshift__',
     '__imod__',
     '__imul__',
     '__index__',
     '__int__',
     '__invert__',
     '__ior__',
     '__ipow__',
     '__irshift__',
     '__isub__',
     '__iter__',
     '__itruediv__',
     '__ixor__',
     '__le__',
     '__long__',
     '__lshift__',
     '__lt__',
     '__mod__',
     '__mul__',
     '__ne__',
     '__neg__',
     '__nonzero__',
     '__oct__',
     '__or__',
     '__pos__',
     '__pow__',
     '__radd__',
     '__rand__',
     '__rdiv__',
     '__rdivmod__',
     '__reduce__',
     '__reduce_ex__',
     '__repr__',
     '__rfloordiv__',
     '__rlshift__',
     '__rmod__',
     '__rmul__',
     '__ror__',
     '__rpow__',
     '__rrshift__',
     '__rshift__',
     '__rsub__',
     '__rtruediv__',
     '__rxor__',
     '__setitem__',
     '__setslice__',
     '__setstate__',
     '__str__',
     '__sub__',
     '__truediv__',
     '__xor__',
     'all',
     'any',
     'argmax',
     'argmin',
     'argsort',
     'astype',
     'byteswap',
     'choose',
     'clip',
     'compress',
     'conj',
     'conjugate',
     'copy',
     'cumprod',
     'cumsum',
     'diagonal',
     'dump',
     'dumps',
     'fill',
     'flatten',
     'getfield',
     'item',
     'itemset',
     'max',
     'mean',
     'min',
     'nonzero',
     'prod',
     'ptp',
     'put',
     'ravel',
     'repeat',
     'reshape',
     'resize',
     'round',
     'searchsorted',
     'setfield',
     'setflags',
     'sort',
     'squeeze',
     'std',
     'sum',
     'swapaxes',
     'take',
     'tofile',
     'tolist',
     'tostring',
     'trace',
     'transpose',
     'var',
     'view']) - numpy_inplace_methods


class SynchronisationError(RuntimeError):
    '''
    Indicates that the GPU and CPU data are in conflict (i.e. both have changed)
    '''
    pass

# These two functions are decorators to be applied to the numpy
# methods. The inplace version sets the cpu data changed flag
# after it has run, the access version only makes sure that
# the data has been synchronised to the CPU.
def gpu_buffered_array_inplace_method(func):
    def new_func(self, *args, **kwds):
        self.sync_to_cpu()
        rval = func(self, *args, **kwds)
        self._cpu_data_changed = True
        return rval
    new_func.__name__ = func.__name__
    new_func.__doc__ = func.__doc__
    return new_func

def gpu_buffered_array_access_method(func):
    def new_func(self, *args, **kwds):
        self.sync_to_cpu()
        rval = func(self, *args, **kwds)
        return rval
    new_func.__name__ = func.__name__
    new_func.__doc__ = func.__doc__
    return new_func


class DecoupledGPUBufferedArray(numpy.ndarray):
    pass


class GPUBufferedArray(numpy.ndarray):
    '''
    An object that automatically manages a numpy array that is stored on the CPU and GPU
    
    The idea is that whenever data is accessed on either the GPU or CPU, this object
    performs any synchronisation operations. So for example if you change a value in the
    array on the CPU and then run a function on the GPU, the data from the CPU side
    needs to be copied to the GPU before calling the GPU function.
    
    The best practice is to use the ``cpu_array`` and ``gpu_array`` or ``gpu_dev_alloc``
    attributes explicitly, and call the ``changed_cpu_data()`` and ``changed_gpu_data()`` methods
    after modifying data. However, the class does its best to be intelligent about
    when changes to data have happened so these arrays can be used in code that is
    not GPU aware to some extent. See implementation notes below for details on when
    this will fail.
    
    **Initialisation**
    
    Initialising with a numpy array creates a ``pycuda.gpuarray.GPUArray`` instance
    to store the data on the GPU. If you want, you can specify a pre-existing
    ``GPUArray`` or ``pycuda.DeviceAllocation`` instance. Initialising with a
    pagelocked numpy array will make copies faster.
    
    **Attributes**
    
    Each of these attributes returns a reference to data on the CPU or GPU (and
    like everything in this class, will synchronise if necessary). Obtaining a
    reference to CPU/GPU data is taken as implying that the data will be
    changed by subsequent code, and so the CPU/GPU data is marked as changed.
    To obtain references without marking the data as changed, use the
    attribute with ``_nomodify`` appended. If you are obtaining and storing
    a reference, you need to explicitly mark it as modified when you make
    a change (see the ``changed_gpu_data()`` and ``changed_cpu_data()`` methods below).
    
    ``cpu_array``, ``cpu_array_nomodify``
        A reference to the array on the CPU.
    ``gpu_array``, ``gpu_array_nomodify``
        If the GPU data is a ``GPUArray``, this will return it, otherwise it will
        throw an error.
    ``gpu_dev_alloc``, ``gpu_dev_alloc_nomodify``
        Return the ``DeviceAllocation`` for the GPU data.
    ``gpu_pointer``, ``gpu_pointer_nomodify``
        Returns an integer pointer to the GPU data.
    
    **Methods**
    
    ``changed_gpu_data()``, ``changed_cpu_data()``
        Mark the data on the GPU or CPU as modified. You need to do this
        explicitly if you obtained and stored a reference to the CPU or GPU
        data.
    ``sync_to_gpu()``, ``sync_to_cpu()``
        Explicitly make sure that the data on the CPU/GPU has been copied to
        the GPU/CPU respectively if they are marked as changed.
    ``sync()``
        Explicitly synchronise the CPU and GPU.
    
    **Implementation**
    
    The implementation is not clever about copying data, it either copies all the data
    or none at all. Internally it stores two flags to indicate whether data on the CPU
    or GPU has changed, which it tries its best to be accurate but cannot guarantee
    (see below), and performs a copy when needed (subsequently updating these flags).
    
    Results of numpy functions, except for a handful of explicitly inplace operators,
    do not mark the data as changed. So doing, for example, ``clip(y,0,1,out=y)``
    will not mark the data in ``y`` as changed.
    '''
    def __new__(subtype, cpu_arr, gpu_arr=None):
        return numpy.array(cpu_arr, copy=False).view(subtype)

    def __init__(self, cpu_arr, gpu_arr=None):
        if gpu_arr is None:
            if DEBUG_BUFFER_CACHE:
                print 'Initialising and copying GPU memory'
            self._gpu_arr = pycuda.gpuarray.to_gpu(self)
            self._gpu_data_changed = False
            self._cpu_data_changed = False
        else:
            self._gpu_arr = gpu_arr
            self._gpu_data_changed = False
            self._cpu_data_changed = True

    def _check_synchronisation(self):
        if not hasattr(self, '_cpu_data_changed'):
            # This happens if you do, for example, y[0,:][:]=1 because the y[0,:] will
            # return an uninitialised GPUBufferedArray. At the moment, we don't
            # handle this case (although perhaps in the case of slicing we could?) so
            # maybe we should throw an error here? The alternative is to try to minimise
            # the impact on code (although it will have decoupled the array from the
            # GPU data, which may not be expected).
            self.__class__ = DecoupledGPUBufferedArray
            self._cpu_data_changed = False
            self._gpu_data_changed = False
            return
        if self._cpu_data_changed and self._gpu_data_changed:
            raise SynchronisationError('GPU and CPU data desynchronised.')

    def sync(self):
        self.sync_to_gpu()
        self.sync_to_cpu()

    def sync_to_gpu(self):
        self._check_synchronisation()
        if self._cpu_data_changed:
            gpu_arr = self._gpu_arr
            if isinstance(gpu_arr, pycuda.gpuarray.GPUArray):
                gpu_arr.set(self)
            else:
                pycuda.driver.memcpy_htod(gpu_arr, self)
            self._cpu_data_changed = False
            if DEBUG_BUFFER_CACHE:
                print 'Synchronised to GPU'

    def sync_to_cpu(self):
        self._check_synchronisation()
        if self._gpu_data_changed:
            gpu_arr = self._gpu_arr
            if isinstance(gpu_arr, pycuda.gpuarray.GPUArray):
                gpu_arr.get(self)
            else:
                pycuda.driver.memcpy_dtoh(self, gpu_arr)
            self._gpu_data_changed = False
            if DEBUG_BUFFER_CACHE:
                print 'Synchronised to CPU'

    def changed_gpu_data(self):
        self._gpu_data_changed = True
        self._check_synchronisation()

    def changed_cpu_data(self):
        self._cpu_data_changed = True
        self._check_synchronisation()

    def get_cpu_array(self, modify=True):
        self.sync_to_cpu()
        if modify: # assume that the user is going to modify the data
            self._cpu_data_changed = True
        return self

    def get_gpu_array(self, modify=True):
        self.sync_to_gpu()
        if modify: # assume that the user is going to modify the data
            self._gpu_data_changed = True
        if isinstance(self._gpu_arr, pycuda.gpuarray.GPUArray):
            return self._gpu_arr
        raise TypeError('GPU buffer is not a GPUArray')

    def get_gpu_dev_alloc(self, modify=True):
        self.sync_to_gpu()
        if modify: # assume that the user is going to modify the data
            self._gpu_data_changed = True
        if isinstance(self._gpu_arr, pycuda.gpuarray.GPUArray):
            return self._gpu_arr.gpudata
        elif isinstance(self._gpu_arr, pycuda.driver.DeviceAllocation):
            return self._gpu_arr
        raise TypeError('gpu_arr should be a DeviceAllocation or GPUArray.')

    def get_gpu_pointer(self):
        return intp(self.gpu_dev_alloc)

    cpu_array = property(fget=get_cpu_array)
    gpu_array = property(fget=get_gpu_array)
    gpu_dev_alloc = property(fget=get_gpu_dev_alloc)
    gpu_pointer = property(fget=get_gpu_pointer)
    cpu_array_nomodify = property(fget=lambda self:self.get_cpu_array(False))
    gpu_array_nomodify = property(fget=lambda self:self.get_gpu_array(False))
    gpu_dev_alloc_nomodify = property(fget=lambda self:self.get_gpu_dev_alloc(False))
    gpu_pointer_nomodify = property(fget=lambda self:self.get_gpu_pointer(False))
    for __name in numpy_inplace_methods:
        exec __name + ' = gpu_buffered_array_inplace_method(numpy.ndarray.' + __name + ')'
    for __name in numpy_access_methods:
        exec __name + ' = gpu_buffered_array_access_method(numpy.ndarray.' + __name + ')'

    def __array_wrap__(self, obj, context=None):
        # normally, __array_wrap__ calls __array_finalize__ to convert the result obj into an object of its own type
        # but here, we don't want that, we want it to be a straight numpy array which doesn't maintain a connection
        # to the GPU array.
        return obj

if __name__ == '__main__':
    import pycuda.autoinit
    from pycuda import driver as drv
    def gpuarrstatus(z):
        return '(cpu_changed=' + str(z._cpu_data_changed) + ', gpu_changed=' + str(z._gpu_data_changed) + ')'
    x = array([1, 2, 3, 4], dtype=numpy.float32)
    print '+ About to initialise GPUBufferedArray'
    y = GPUBufferedArray(x)
    print '- Initialised GPUBufferedArray', gpuarrstatus(y)
    print '+ About to set item on CPU array', gpuarrstatus(y)
    y[3] = 5
    print '- Set item on CPU array', gpuarrstatus(y)
    mod = drv.SourceModule('''
    __global__ void doubleit(float *y)
    {
        int i = threadIdx.x;
        y[i] *= 2.0;
    }
    ''')
    doubleit = mod.get_function("doubleit")
    print '+ About to call GPU function', gpuarrstatus(y)
    doubleit(y.gpu_array, block=(4, 1, 1))
    print '- Called GPU function', gpuarrstatus(y)
    print '+ About to print CPU array', gpuarrstatus(y)
    print y
    print '- Printed CPU array', gpuarrstatus(y)



'''
TODO: for some equations should we first reduce and then solve rather than
      solving and then reducing? probably...
'''

from sympy import Symbol, solve
import re

def sides(line):
    m = re.match(r'([^=]*)=([^e]*)', line)
    lhs, rhs = m.groups()
    return lhs.strip(), rhs.strip()

def replace_differentials(expr, replacements):
    while True:
        m = re.search(r'd\s*(\d*)\s*(\w+)\s*/\s*d\s*t\s*(\d*)', expr)
        if not m: break
        n1, var, n2 = m.groups()
        if n1 != n2:
            raise ValueError('Inconsistent indices in differential')
        if n1 == '':
            n1 = '1'
        n = int(n1)
        newsym = '_diff_' + var + '_' + n1
        replacements[newsym] = (var, n, expr[m.start():m.end()])
        expr = expr[:m.start()] + newsym + expr[m.end():]
    return expr

def parse(eqs):
    print 'Original equations:'
    print '\n'.join(line.strip() for line in eqs.strip().split('\n'))

    lines = eqs.split('\n')
    lines = [line.strip() for line in lines]
    lines = [line for line in lines if line]
    lines = map(sides, lines)
    replacements = {}
    eqstosolve = []
    print
    print 'Equations after symbol replacement:'
    for lhs, rhs in lines:
        lhs = replace_differentials(lhs, replacements)
        rhs = replace_differentials(rhs, replacements)
        eqstosolve.append('(%s)-(%s)' % (lhs, rhs))
        print lhs, '=', rhs
    for k, (v, n, r) in replacements.iteritems():
        print 'Symbol', k, 'defines order', n, 'differential of', v, ':', r
    # Now solve this system of equations for the differentials
    symbolset = dict((k, Symbol(k)) for k in replacements.keys())
    ns = symbolset.copy()
    ns.update(dict((v, Symbol(v)) for v, _, _ in replacements.values()))
    eqstosolve = [eval(eq, ns) for eq in eqstosolve]
    solutions = solve(eqstosolve, *symbolset.values())
    print
    print 'Solutions:'
    for k, v in solutions.items():
        print k, '=', v
    # Now reduce Nth order equations to  first order ones
    reducedeqs = {}
    print
    print 'Reduced set of equations:'
    for diffeqsymbol, diffeqvalue in solutions.items():
        var, n, r = replacements[str(diffeqsymbol)]
        for i in xrange(n):
            if i == 0:
                newvar = var
            else:
                newvar = '_rdiff_%s_%d' % (var, i)
            if i == n - 1:
                neweq = diffeqvalue
            else:
                neweq = '_rdiff_%s_%d' % (var, i + 1)
            reducedeqs[newvar] = neweq
            print 'd%s/dt = %s' % (newvar, neweq)

if __name__ == '__main__':
#    eqs = '''
#    a * d2x/dt2 + b * d2y/dt2 = c
#    d * d2x/dt2 + e * d2y/dt2 = f
#    '''
    eqs = '''
    3 * d2x/dt2 + 2 * d2y/dt2 = 6*x
    4 * d2x/dt2 + 1 * d2y/dt2 = 5*y
    '''
    parse(eqs)



'''
This package maintains multiple patches for scipy.sparse.lil_matrix for the
various different versions of scipy. It chooses dynamically at run time which
version to use.
'''
import scipy, warnings, os, imp
import scipy.sparse

__all__ = ['lil_matrix', 'spmatrix']

default = '0_7_1'
table = {
    '0.9.0':     '0_7_1',
    '0.8.0':     '0_7_1',
    '0.7.2':     '0_7_1',
    '0.7.1':     '0_7_1',
    '0.7.1rc3':  '0_7_1',
    '0.7.1rc2':  '0_7_1',
    '0.7.1rc1':  '0_7_1',
    '0.7.0':     '0_7_0',
    '0.7.0rc2':  '0_7_0',
    '0.7.0b1':   '0_7_0',
    '0.6.0':     '0_6_0',
    }

if scipy.__version__ in table:
    modulename = table[scipy.__version__]
else:
    for i in xrange(1, len(scipy.__version__)):
        v = scipy.__version__[:-i]
        if v in table:
            modulename = table[v]
            break
    else:
        modulename = default
        warnings.warn("Couldn't find matching sparse matrix patch for scipy version %s, but in most cases this shouldn't be a problem." % scipy.__version__)

module = __import__(modulename, globals(), locals(), [], -1)
lil_matrix = module.lil_matrix
spmatrix = scipy.sparse.spmatrix

if __name__ == '__main__':
    x = lil_matrix((5, 5))
    x[2, 3] = 5
    print x.todense()




'''
Synapse models for Brian (no plasticity here).
'''
from brian.equations import *
from brian.units import check_units, second, amp, siemens
from brian.membrane_equations import Current

__all__ = ['exp_current', 'alpha_current', 'biexp_current', \
         'exp_conductance', 'alpha_conductance', 'biexp_conductance', \
         'exp_synapse', 'alpha_synapse', 'biexp_synapse']

__credits__ = dict(author='Romain Brette (brette@di.ens.fr)',
                 date='April 2008')

# -----------------
# Synaptic currents
# -----------------
@check_units(tau=second)
def exp_current(input, tau, current_name=None, unit=amp):
    '''
    Exponential synaptic current.
    input = name of input variable (where presynaptic spikes act).
    current_name = name of current variable
    '''
    current_name = current_name or unique_id()
    current = Current() + exp_synapse(input, tau, unit, current_name)
    current.set_current_name(current_name)
    return current

@check_units(tau=second)
def alpha_current(input, tau, current_name=None, unit=amp):
    '''
    Alpha synaptic current.
    current_name = name of current variable
    '''
    current_name = current_name or unique_id()
    current = Current() + alpha_synapse(input, tau, unit, current_name)
    current.set_current_name(current_name)
    return current

@check_units(tau1=second, tau2=second)
def biexp_current(input, tau1, tau2, current_name=None, unit=amp):
    '''
    Biexponential synaptic current.
    current_name = name of current variable
    '''
    current_name = current_name or unique_id()
    current = Current() + biexp_synapse(input, tau1, tau2, unit, current_name)
    current.set_current_name(current_name)
    return current

# ---------------------
# Synaptic conductances
# ---------------------
@check_units(tau=second)
def exp_conductance(input, E, tau, conductance_name=None, unit=siemens):
    '''
    Exponential synaptic conductance.
    conductance_name = name of conductance variable
    E = synaptic reversal potential
    '''
    conductance_name = conductance_name or unique_id()
    return Current('I=g*(E-vm): amp', I=input + '_current', g=conductance_name, E=E) + \
           exp_synapse(input, tau, unit, conductance_name)

@check_units(tau=second)
def alpha_conductance(input, E, tau, conductance_name=None, unit=siemens):
    '''
    Alpha synaptic conductance.
    conductance_name = name of conductance variable
    E = synaptic reversal potential
    '''
    conductance_name = conductance_name or unique_id()
    return Current('I=g*(E-vm): amp', I=input + '_current', g=conductance_name, E=E) + \
           alpha_synapse(input, tau, unit, conductance_name)

@check_units(tau1=second, tau2=second)
def biexp_conductance(input, E, tau1, tau2, conductance_name=None, unit=siemens):
    '''
    Exponential synaptic conductance.
    conductance_name = name of conductance variable
    E = synaptic reversal potential
    '''
    conductance_name = conductance_name or unique_id()
    return Current('I=g*(E-vm): amp', I=input + '_current', g=conductance_name, E=E) + \
           biexp_synapse(input, tau1, tau2, unit, conductance_name)

# ---------------
# Synaptic inputs
# ---------------
@check_units(tau=second)
def exp_synapse(input, tau, unit, output=None):
    '''
    Exponentially decaying synaptic current/conductance:
    g(t)=exp(-t/tau)
    output = output variable name (plugged into the membrane equation).
    input = input variable name (where spikes are received).
    '''
    if output is None:
        output = input + '_out'
    return Equations('''
    dx/dt = -x*invtau    : unit
    y=x''', x=output, y=input, unit=unit, invtau=1. / tau)

@check_units(tau=second)
def alpha_synapse(input, tau, unit, output=None):
    '''
    Alpha synaptic current/conductance:
    g(t)=(t/tau)*exp(1-t/tau)
    output = output variable name (plugged into the membrane equation).
    input = input variable name (where spikes are received).
    The peak is 1 at time t=tau.
    '''
    if output is None:
        output = input + '_out'
    return Equations('''
    dx/dt = (y-x)*invtau : unit 
    dy/dt = -y*invtau    : unit
    ''', x=output, y=input, unit=unit, invtau=1. / tau)

@check_units(tau1=second, tau2=second)
def biexp_synapse(input, tau1, tau2, unit, output=None):
    '''
    Biexponential synaptic current/conductance:
    g(t)=(tau2/(tau2-tau1))*(exp(-t/tau1)-exp(-t/tau2))
    output = output variable name (plugged into the membrane equation).
    input = input variable name (where spikes are received).
    The peak is 1 at time t=tau1*tau2/(tau2-tau1)*log(tau2/tau1) 
    '''
    if output is None:
        output = input + '_out'
    invpeak = (tau2 / tau1) ** (tau1 / (tau2 - tau1))
    return Equations('''
    dx/dt = (invpeak*y-x)*invtau1 : unit
    dy/dt = -y*invtau2            : unit
    ''', x=output, y=input, unit=unit, invtau1=1. / tau1, invtau2=1. / tau2, invpeak=invpeak)



from brian import *
from numpy import *

__all__ = ['dB', 'dB_type', 'dB_error', 'gain']

class dB_error(ValueError):
    '''
    Error raised when values in dB are used inconsistently with other units.
    '''
    pass

class dB_type(float64):
    '''
    The type of values in dB.
    
    dB values are assumed to be RMS dB SPL assuming that the sound source is
    measured in Pascals.
    '''
    def __str__(self):
        return str(float(self))+'*dB'
    def __repr__(self):
        return repr(float(self))+'*dB'
    def __mul__(self, other):
        if isinstance(other, dB_type):
            raise dB_error('Cannot multiply dB by dB')
        return dB_type(float(self)*other)
    __rmul__ = __mul__
    def __div__(self, other):
        if isinstance(other, dB_type):
            raise dB_error('Cannot divide dB by dB')
        return dB_type(float(self)/other)
    __truediv__ = __div__
    def __rdiv__(self, other):
        if isinstance(other, dB_type):
            raise dB_error('Cannot divide dB by dB')
        return dB_type(other/float(self))
    __rtruediv__ = __rdiv__
    def __add__(self, other):
        if not isinstance(other, dB_type):
            raise dB_error('Cannot add a dB object to a non-dB object')
        return dB_type(float(self)+float(other))
    __radd__ = __add__
    def __sub__(self, other):
        if not isinstance(other, dB_type):
            raise dB_error('Cannot subtract a dB object from a non-dB object')
        return dB_type(float(self)-float(other))
    def __rsub__(self, other):
        if not isinstance(other, dB_type):
            raise dB_error('Cannot subtract a dB object from a non-dB object')
        return dB_type(float(other)-float(self))
    def __neg__(self):
        return dB_type(-float(self))
    def __pos__(self):
        return self
    def __abs__(self):
        return dB_type(abs(float(self)))
    def __pow__(self, other):
        raise dB_error('Cannot take powers with dB')
    __rpow__ = __pow__
    def __lt__(self, other):
        if not isinstance(other, dB_type):
            raise dB_error('Can only compare with another dB')
        return float(self)<float(other)
    def __le__(self, other):
        if not isinstance(other, dB_type):
            raise dB_error('Can only compare with another dB')
        return float(self)<=float(other)
    def __gt__(self, other):
        if not isinstance(other, dB_type):
            raise dB_error('Can only compare with another dB')
        return float(self)>float(other)
    def __ge__(self, other):
        if not isinstance(other, dB_type):
            raise dB_error('Can only compare with another dB')
        return float(self)>=float(other)
    def __eq__(self, other):
        if not isinstance(other, dB_type):
            raise dB_error('Can only compare with another dB')
        return float(self)==float(other)
    def __ne__(self, other):
        if not isinstance(other, dB_type):
            raise dB_error('Can only compare with another dB')
        return float(self)!=float(other)
    def __reduce__(self):
        return (dB_type, (float(self),))
    def gain(self):
        return 10**(float(self)/20.0)


dB = dB_type(1.0)

def gain(level):
    '''
    Returns the gain factor associated to a level in dB.
    
    The formula is:
    
    gain = 10**(level/20.0)
    '''
    if not isinstance(level, dB_type):
        raise dB_error('Level must be in dB')
    return level.gain()



################################################################################
################## NO DELAYS ###################################################
################################################################################

################## SPARSE ######################################################

propagate_weave_code_sparse_vars = ['sv', 'rowinds', 'datas', 'spikes', 'nspikes']
propagate_weave_code_sparse = '''
    for(int j=0;j<nspikes;j++)
    {
        PyObject* _rowind = rowinds[j];
        PyArrayObject* _row = convert_to_numpy(_rowind, "row");
        conversion_numpy_check_type(_row, PyArray_LONG, "row");
        conversion_numpy_check_size(_row, 1, "row");
        //blitz::Array<long,1> row = convert_to_blitz<long,1>(_row,"row");
        long* row = (long*)_row->data;
        PyObject* _datasj = datas[j];
        PyArrayObject* _data = convert_to_numpy(_datasj, "data");
        conversion_numpy_check_type(_data, PyArray_DOUBLE, "data");
        conversion_numpy_check_size(_data, 1, "data");
        //blitz::Array<double,1> data = convert_to_blitz<double,1>(_data,"data");
        double* data = (double*)_data->data;
        //int m = row.numElements();
        int m = _row->dimensions[0];
        for(int k=0;k<m;k++)
            //sv(row(k)) += data(k);
            sv[row[k]] += data[k];
        Py_DECREF(_rowind);
        Py_DECREF(_datasj);
    }
    '''
    
propagate_weave_code_sparse_modulation_vars = ['sv', 'sv_pre', 'rowinds', 'datas', 'spikes', 'nspikes']
propagate_weave_code_sparse_modulation = '''
    for(int j=0;j<nspikes;j++)
    {
        PyObject* _rowind = rowinds[j];
        PyArrayObject* _row = convert_to_numpy(_rowind, "row");
        conversion_numpy_check_type(_row, PyArray_LONG, "row");
        conversion_numpy_check_size(_row, 1, "row");
        //blitz::Array<long,1> row = convert_to_blitz<long,1>(_row,"row");
        long* row = (long*)_row->data;
        PyObject* _datasj = datas[j];
        PyArrayObject* _data = convert_to_numpy(_datasj, "data");
        conversion_numpy_check_type(_data, PyArray_DOUBLE, "data");
        conversion_numpy_check_size(_data, 1, "data");
        //blitz::Array<double,1> data = convert_to_blitz<double,1>(_data,"data");
        double* data = (double*)_data->data;
        //int m = row.numElements();
        int m = _row->dimensions[0];
        //double mod = sv_pre(spikes(j));
        double mod = sv_pre[spikes[j]];
        for(int k=0;k<m;k++)
            //sv(row(k)) += data(k)*mod;
            sv[row[k]] += data[k]*mod;
        Py_DECREF(_rowind);
        Py_DECREF(_datasj);
    }
    '''

################## DENSE #######################################################

propagate_weave_code_dense_vars = ['sv', 'spikes', 'nspikes', 'N', 'rows']
propagate_weave_code_dense = '''
    for(int j=0;j<nspikes;j++)
    {
        PyObject* _rowsj = rows[j];
        PyArrayObject* _row = convert_to_numpy(_rowsj, "row");
        conversion_numpy_check_type(_row, PyArray_DOUBLE, "row");
        conversion_numpy_check_size(_row, 1, "row");
        //blitz::Array<double,1> row = convert_to_blitz<double,1>(_row,"row");
        double *row = (double *)_row->data;
        for(int k=0;k<N;k++)
            //sv(k) += row(k);
            sv[k] += row[k];
        Py_DECREF(_rowsj);
    }
    '''

propagate_weave_code_dense_modulation_vars = ['sv', 'sv_pre', 'spikes', 'nspikes', 'N', 'rows']
propagate_weave_code_dense_modulation = '''
    for(int j=0;j<nspikes;j++)
    {
        PyObject* _rowsj = rows[j];
        PyArrayObject* _row = convert_to_numpy(_rowsj, "row");
        conversion_numpy_check_type(_row, PyArray_DOUBLE, "row");
        conversion_numpy_check_size(_row, 1, "row");
        //blitz::Array<double,1> row = convert_to_blitz<double,1>(_row,"row");
        //double mod = sv_pre(spikes(j));
        double *row = (double *)_row->data;
        double mod = sv_pre[spikes[j]];
        for(int k=0;k<N;k++)
            sv[k] += row[k]*mod;
            //sv(k) += row(k)*mod;
        Py_DECREF(_rowsj);
    }
    '''

################################################################################
################## DELAYS ######################################################
################################################################################

################## SPARSE ######################################################

delay_propagate_weave_code_sparse_vars = [
    'rowinds', 'datas', 'spikes', 'nspikes', 'dvecrows', 'dr', 'cdi', 'idt',
    'md']
delay_propagate_weave_code_sparse = """
    for(int j=0;j<nspikes;j++)
    {
        PyObject* _rowind = rowinds[j];
        PyArrayObject* _row = convert_to_numpy(_rowind, "row");
        conversion_numpy_check_type(_row, PyArray_LONG, "row");
        conversion_numpy_check_size(_row, 1, "row");
        blitz::Array<long,1> row = convert_to_blitz<long,1>(_row,"row");
        PyObject* _datasj = datas[j];
        PyArrayObject* _data = convert_to_numpy(_datasj, "data");
        conversion_numpy_check_type(_data, PyArray_DOUBLE, "data");
        conversion_numpy_check_size(_data, 1, "data");
        blitz::Array<double,1> data = convert_to_blitz<double,1>(_data,"data");
        PyObject* _dvecrowsj = dvecrows[j];
        PyArrayObject* _dvecrow = convert_to_numpy(_dvecrowsj, "dvecrow");
        conversion_numpy_check_type(_dvecrow, PyArray_DOUBLE, "dvecrow");
        conversion_numpy_check_size(_dvecrow, 1, "dvecrow");
        blitz::Array<double,1> dvecrow = convert_to_blitz<double,1>(_dvecrow,"dvecrow");
        int m = row.numElements();
        for(int k=0;k<m;k++)
        {
            dr((cdi+(int)(idt*dvecrow(k)))%md, (int)row(k)) += data(k);
        }
        Py_DECREF(_rowind);
        Py_DECREF(_datasj);
        Py_DECREF(_dvecrowsj);
    }
    """
    
delay_propagate_weave_code_sparse_modulation_vars = [
    'sv_pre', 'rowinds', 'datas', 'spikes', 'nspikes', 'dvecrows', 'dr', 'cdi',
    'idt', 'md']
delay_propagate_weave_code_sparse_modulation = """
    for(int j=0;j<nspikes;j++)
    {
        PyObject* _rowind = rowinds[j];
        PyArrayObject* _row = convert_to_numpy(_rowind, "row");
        conversion_numpy_check_type(_row, PyArray_LONG, "row");
        conversion_numpy_check_size(_row, 1, "row");
        blitz::Array<long,1> row = convert_to_blitz<long,1>(_row,"row");
        PyObject* _datasj = datas[j];
        PyArrayObject* _data = convert_to_numpy(_datasj, "data");
        conversion_numpy_check_type(_data, PyArray_DOUBLE, "data");
        conversion_numpy_check_size(_data, 1, "data");
        blitz::Array<double,1> data = convert_to_blitz<double,1>(_data,"data");
        PyObject* _dvecrowsj = dvecrows[j];
        PyArrayObject* _dvecrow = convert_to_numpy(_dvecrowsj, "dvecrow");
        conversion_numpy_check_type(_dvecrow, PyArray_DOUBLE, "dvecrow");
        conversion_numpy_check_size(_dvecrow, 1, "dvecrow");
        blitz::Array<double,1> dvecrow = convert_to_blitz<double,1>(_dvecrow,"dvecrow");
        int m = row.numElements();
        double mod = sv_pre(spikes(j));
        for(int k=0;k<m;k++)
        {
            dr((cdi+(int)(idt*dvecrow(k)))%md, (int)row(k)) += data(k)*mod;
        }
        Py_DECREF(_rowind);
        Py_DECREF(_datasj);
        Py_DECREF(_dvecrowsj);
    }
    """

################## DENSE #######################################################

delay_propagate_weave_code_dense_vars = [
    'spikes', 'nspikes', 'N', 'rows', 'dr', 'cdi', 'idt', 'md', 'dvecrows']
delay_propagate_weave_code_dense = """
    for(int j=0;j<nspikes;j++)
    {
        PyObject* _rowsj = rows[j];
        PyArrayObject* _row = convert_to_numpy(_rowsj, "row");
        conversion_numpy_check_type(_row, PyArray_DOUBLE, "row");
        conversion_numpy_check_size(_row, 1, "row");
        blitz::Array<double,1> row = convert_to_blitz<double,1>(_row,"row");
        PyObject* _dvecrowsj = dvecrows[j];
        PyArrayObject* _dvecrow = convert_to_numpy(_dvecrowsj, "dvecrow");
        conversion_numpy_check_type(_dvecrow, PyArray_DOUBLE, "dvecrow");
        conversion_numpy_check_size(_dvecrow, 1, "dvecrow");
        blitz::Array<double,1> dvecrow = convert_to_blitz<double,1>(_dvecrow,"dvecrow");
        for(int k=0;k<N;k++)
            dr((cdi+(int)(idt*dvecrow(k)))%md, k) += row(k);
        Py_DECREF(_rowsj);
        Py_DECREF(_dvecrowsj);
    }
    """

delay_propagate_weave_code_dense_modulation_vars = [
    'sv_pre', 'spikes', 'nspikes', 'N', 'rows', 'dr', 'cdi', 'idt', 'md',
    'dvecrows']
delay_propagate_weave_code_dense_modulation = """
    for(int j=0;j<nspikes;j++)
    {
        PyObject* _rowsj = rows[j];
        PyArrayObject* _row = convert_to_numpy(_rowsj, "row");
        conversion_numpy_check_type(_row, PyArray_DOUBLE, "row");
        conversion_numpy_check_size(_row, 1, "row");
        blitz::Array<double,1> row = convert_to_blitz<double,1>(_row,"row");
        PyObject* _dvecrowsj = dvecrows[j];
        PyArrayObject* _dvecrow = convert_to_numpy(_dvecrowsj, "dvecrow");
        conversion_numpy_check_type(_dvecrow, PyArray_DOUBLE, "dvecrow");
        conversion_numpy_check_size(_dvecrow, 1, "dvecrow");
        blitz::Array<double,1> dvecrow = convert_to_blitz<double,1>(_dvecrow,"dvecrow");
        double mod = sv_pre(spikes(j));
        for(int k=0;k<N;k++)
            dr((cdi+(int)(idt*dvecrow(k)))%md, k) += row(k)*mod;
        Py_DECREF(_rowsj);
        Py_DECREF(_dvecrowsj);
    }
    """



'''
Implementation of the basic model (no speech recognition, no learning) 
described in:
Gutig and Sompolinsky (2009): "Time-Warp-Invariant Neuronal Processing"
PLoS Biology, Vol. 7 (7), e1000141
'''
from brian import *

class TimeWarpModel(object):
    '''
    A simple neuron model for testing the "time-warp invariance" with 
    conductance based or current based synapses. The neuron receives balanced
    excitatory and inhibitory input from a random spike train. The same spike
    train can be fed into the model with different time warps.
    ''' 
    def __init__(self, conductance_based=True):
        '''
        Create a new model with conductance based or current based synapses
        '''
        # Model parameters
        E_e = 5
        E_i = -1
        E_L = 0
        g_L = 1/(100.0*msecond)
        tau_syn = 1*ms
        N_ex = 250
        N_inh = 250
        self.N = N_ex + N_inh
                
        # Equations
        if conductance_based:
            eqs = '''
                dV/dt = -(V - E_L) * g_L - I_syn  : 1
                I_syn = I_ge + I_gi  : second**-1
                I_ge = (V - E_e) * g_e : second**-1
                I_gi = (V - E_i) * g_i : second**-1
                dg_e/dt = -g_e/tau_syn : second**-1
                dg_i/dt = -g_i/tau_syn : second**-1
                '''
        else:
            eqs = '''
                dV/dt = -(V - E_L) * g_L - I_syn  : 1
                I_syn = -5 * g_e + g_i : second**-1
                dg_e/dt = -g_e/tau_syn : second**-1
                dg_i/dt = -g_i/tau_syn : second**-1
                '''        
        
        # for simpler voltage traces: no spiking
        neuron = NeuronGroup(1, model=eqs, threshold=None)
            
        # every input neuron fires once in a random interval
        self.unwarped_spiketimes = [(i, t * 250 * ms) for i, t in 
                                    zip(range(0, self.N), rand(self.N))]
        
        # final spiketimes will be set in the run function
        self.input = SpikeGeneratorGroup(self.N, [])
                
        e_input = self.input.subgroup(N_ex)
        i_input = self.input.subgroup(N_inh)        
        e_conn = Connection(e_input, neuron, 'g_e',
                                  weight=6 / (N_ex * tau_syn))
        i_conn = Connection(i_input, neuron, 'g_i',
                                  weight=5 * 6 / (N_ex * tau_syn)) 
        
        # record membrane potential
        self.monitor = StateMonitor(neuron, varname='V', record=True)

        # putting everything together
        self.net = Network(neuron, self.input, e_conn, i_conn, self.monitor)

    def run(self, beta=1.0):
        ''' 
        Run the network with the original spike train warped by a certain factor
        beta. Beta > 1 corresponds to an extended and beta < 1 to a shrinked
        input spike train.
        '''
        self.net.reinit()
        
        #warp spike train in time        
        self.input.spiketimes = [(i, beta*t) 
                                 for i, t in self.unwarped_spiketimes]        
        self.net.run(beta * 250*ms)
        
        #Return the voltage trace
        return (self.monitor.times, self.monitor[0])


if __name__ == '__main__':
    cond_model = TimeWarpModel(True)
    curr_model = TimeWarpModel(False)
    N = cond_model.N    
    
    # #########################################################################
    # Reproduce Fig. 2 from Gütig and Sompolinsky (2009)
    # #########################################################################
    beta = 2.0
    times1, v1 = cond_model.run(beta=1.0)
    times2, v2 = cond_model.run(beta=beta)
    maxtime = 250 * beta
    subplot(4, 1, 1)
    (neurons, times) = zip(*cond_model.unwarped_spiketimes)
    plot(array(times) / ms, neurons, 'g.')
    axis([0, maxtime, 0, N])
    xticks([])
    yticks([])
    title('Time-warp-invariant voltage traces (conductance-based)')
    
    subplot(4, 1, 2)
    plot(times1 / ms, v1, 'g')
    axis([0, maxtime, -1.5, 1.5])
    xticks([])
    yticks([])
    
    subplot(4, 1, 3)
    plot(array(times) * beta / ms, neurons, 'b.')
    axis([0, maxtime, 0, N])
    xticks([])
    yticks([1, 500])
    
    subplot(4, 1, 4)
    plot(times2 / ms, v2, 'b')
    plot(times1 / ms * beta, v1, 'g')
    axis([0, maxtime, -1.5, 1.5])
    xlabel('Time (ms)')
    xticks([0, 250, 500])
    yticks([-1, 1])
    show()
    # #########################################################################
    # Reproduce Fig. 3(C) from Gütig and Sompolinsky (2009), but for random
    # spike trains and not in a speech recognition task
    # #########################################################################

    betas = arange(0.2, 3.1, 0.1)
    #betas = array([1.0, 2.0])
    cond_results = []
    curr_results = []
    for beta in betas:
        print 'Testing warp factor %.1f' % beta
        cond_results.append(cond_model.run(beta))
        curr_results.append(curr_model.run(beta))
    
    figure()
    colors = mpl.cm.gist_earth((betas - betas[0]) / (betas[-1] - betas[0]))
    lookup = dict(zip(betas, colors))
    for beta, cond_result, curr_result in zip(betas, cond_results, 
                                              curr_results):
        times_cond, v_cond = cond_result
        times_curr, v_curr = curr_result 
        subplot(1,2,1)
        plot(times_cond / ms / beta, v_cond, color=lookup[beta])
        axis([0, 250, -1.5, 1.5])
        subplot(1,2,2)
        plot(times_curr / ms / beta, v_curr, color=lookup[beta])
        axis([0, 250, -1.5, 1.5])
    subplot(1,2,1)
    title('conductance based')
    subplot(1,2,2)  
    title('current based')
    show()